
-- file: adc_interface_exdes.vhd
--
-- DISCLAIMER OF LIABILITY
--
-- This file contains proprietary and confidential information of
-- Xilinx, Inc. ("Xilinx"), that is distributed under a license
-- from Xilinx, and may be used, copied and/or disclosed only
-- pursuant to the terms of a valid license agreement with Xilinx.
--
-- XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION
-- ("MATERIALS") "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
-- EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING WITHOUT
-- LIMITATION, ANY WARRANTY WITH RESPECT TO NONINFRINGEMENT,
-- MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE. Xilinx
-- does not warrant that functions included in the Materials will
-- meet the requirements of Licensee, or that the operation of the
-- Materials will be uninterrupted or error-free, or that defects
-- in the Materials will be corrected. Furthermore, Xilinx does
-- not warrant or make any representations regarding use, or the
-- results of the use, of the Materials in terms of correctness,
-- accuracy, reliability or otherwise.
--
-- Xilinx products are not designed or intended to be fail-safe,
-- or for use in any application requiring fail-safe performance,
-- such as life-support or safety devices or systems, Class III
-- medical devices, nuclear facilities, applications related to
-- the deployment of airbags, or any other applications that could
-- lead to death, personal injury or severe property or
-- environmental damage (individually and collectively, "critical
-- applications"). Customer assumes the sole risk and liability
-- of any use of Xilinx products in critical applications,
-- subject only to applicable laws and regulations governing
-- limitations on product liability.
--
-- Copyright 2009 Xilinx, Inc.
-- All rights reserved.
--
-- This disclaimer and copyright notice must be retained as part
-- of this file at all times.
--

------------------------------------------------------------------------------
-- SelectIO wizard example design
------------------------------------------------------------------------------
-- This example design instantiates the IO circuitry
------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

library unisim;
use unisim.vcomponents.all;

entity adc_interface_exdes is
generic (
  -- Clock -> q modeling delay
  TCQ        : time    := 100 ps;
  -- width of the data for the system
  sys_w      : integer := 6;
  -- width of the data for the device
  dev_w      : integer := 12;
  -- width of the address for the memory
  add_w      : integer := 10;
  -- depth of the memory
  add_l      : integer := 1024);
port (
  -- Memory interface
  ADDRESS_USER             : in    std_logic_vector(add_w-1 downto 0);
  DATA_IN_USER             : in    std_logic_vector(dev_w-1 downto 0);
  DATA_OUT_USER            : out   std_logic_vector(dev_w-1 downto 0);
  ENABLE_USER              : in    std_logic;
  WRITE_USER               : in    std_logic;
  ENABLE_DEVICE            : in    std_logic;
  -- From the system into the device
  DATA_IN_FROM_PINS_P      : in    std_logic_vector(sys_w-1 downto 0);
  DATA_IN_FROM_PINS_N      : in    std_logic_vector(sys_w-1 downto 0);
  CLK_IN_P                 : in    std_logic;
  CLK_IN_N                 : in    std_logic;
  CLK_RESET                : in    std_logic;
  IO_RESET                 : in    std_logic);
end adc_interface_exdes;

architecture xilinx of adc_interface_exdes is

component adc_interface is
generic
 (-- width of the data for the system
  sys_w       : integer := 6;
  -- width of the data for the device
  dev_w       : integer := 12);
port
 (
  -- From the system into the device
  DATA_IN_FROM_PINS_P     : in    std_logic_vector(sys_w-1 downto 0);
  DATA_IN_FROM_PINS_N     : in    std_logic_vector(sys_w-1 downto 0);
  DATA_IN_TO_DEVICE       : out   std_logic_vector(dev_w-1 downto 0);
  CLK_IN_P                : in    std_logic;
  CLK_IN_N                : in    std_logic;
  CLK_OUT                 : out   std_logic;
  CLK_RESET               : in    std_logic;
  IO_RESET                : in    std_logic);
end component;

   -- connection between ram and io circuit
   signal DATA_IN_TO_DEVICE    : std_logic_vector(dev_w-1 downto 0);
   signal write_device         : std_logic :=  '1';
   signal address_device       : std_logic_vector(add_w-1 downto 0) := (others=>'0');
   type ram_type is array (0 to add_l-1) of std_logic_vector(dev_w-1 downto 0);
   shared variable ram_array   : ram_type;
   signal CLK_OUT              : std_logic;
begin


   -- Infer a dual port memory
   process (CLK_OUT) begin
   if (CLK_OUT = '1' and CLK_OUT'event) then
     if (ENABLE_USER = '1') then
       if (WRITE_USER = '1') then
         ram_array(conv_integer(ADDRESS_USER)) := DATA_IN_USER;
       end if;
       DATA_OUT_USER <= ram_array(conv_integer(ADDRESS_USER))  after TCQ;
     end if;
   end if;
   end process;

   process (CLK_OUT) begin
   if (CLK_OUT = '1' and CLK_OUT'event) then
     if (ENABLE_DEVICE = '1') then
       if (write_device = '1') then
         ram_array(conv_integer(address_device)) := DATA_IN_TO_DEVICE;
       end if;
     end if;
   end if;
   end process;

   -- auto-increment access into the RAM on the IO side
   process (CLK_OUT) begin
   if (CLK_OUT = '1' and CLK_OUT'event) then
     if (IO_RESET = '1') then
       address_device <=  (others=>'0') after TCQ;
     elsif (ENABLE_DEVICE = '1') then
       address_device <= address_device + 1 after TCQ;
     end if;
   end if;
   end process;

   -- Instantiate the IO design
   io_inst : adc_interface
   port map
   (
    -- From the system into the device
    DATA_IN_FROM_PINS_P     => DATA_IN_FROM_PINS_P,
    DATA_IN_FROM_PINS_N     => DATA_IN_FROM_PINS_N,
    DATA_IN_TO_DEVICE       => DATA_IN_TO_DEVICE,

    CLK_IN_P                => CLK_IN_P,
    CLK_IN_N                => CLK_IN_N,
    CLK_OUT                 => CLK_OUT,
    CLK_RESET               => CLK_RESET,
    IO_RESET                => IO_RESET);
end xilinx;



