----------------------------------------------------------------------------------
-- Company: Bluechip Technology
-- Engineer: Istvan Nagy, buenos@freemail.hu
-- 
-- Create Date:    15:58:13 05/30/2010 
-- Design Name: 
-- Module Name:    op2p - Behavioral 
-- Project Name: 
-- Target Devices: Xilinx Spartan-6 LXT, XC6SLX45T or other
-- Tool versions: ISE 12.1
--
-- Dependencies: 
--
-- Revision 0.01 - File Created
-- Additional Comments: 
--
-- Description: Open Peer tp Peer Interface, Wishbone to Aurora Bridge (OP2P).
--   This is only one port. In a flug-in board or blade server we would need an FPGA
--   as the backplane bridge. The OP2P interface is optimized for mesh backplane
--   topologies. The backplane bridge nees a separate port for each backplane link.
--   There has to be a higher level scheduler logic or a soft processor to implement it.
--   To support bigger than purnum^2 elements in a mesh, the bridge has to implement
--   packet forwarding. This way not every node will be directly conenctod to every other.
--   It may be obvious that the forwarding involves 2 ports and some interaction from
--   the processor. The backplane bridge would normally consist of 4 ports, a host interface,
--   a local memory buffer interface (1 DRAM chip on-board) and a local-host (soft proc or
--   statemachine-type logic)
--
-- Protocol: 
-- The lower layers are based on the Xilinx Aurora protocol, and genersted by the Xilinx 
-- Coregenerator tool and instantiated in this file. The higher layers are implemented in 
-- this file.
-- All transactions are peer to peer, non-transparent, and implementing buffer-copy from 
-- devive/address to device/address.
-- Transmit can be initiated by localhost (it writes into local memory, then sets a registerbit 
-- accessed through the aurora_config_wb bus) or by arrived read request packet from remote device.
-- What happens dusring transmit: The IP reads the data from the local buffer, and sends it to the 
-- specified target (destination), then generates an interrupt. Receive: when receiving a packet, 
-- the aurora IP writes it into the memory (address is specified in the packet), then generates
-- an interrupt to the local host. So, all transfersa are directly between the memory and the 
-- aurora link, the host can not write the data into the aurora IP.
--
-- Packet format: Header + payload data. Max 2kBytes in one packet. If the request is tergating
--  a larger than 2kBytes memory area, then the scheduler has to divide the data into 2kB packets.
--  Header: 
--    1st DW: source ID (16bit 31:16), destination ID (16bit 15:0)
--    2nd DW: target address (32bit)
--    3rd DW: source address (32bit)
--    4th DW: byte count(16bit: 1-64k 31:16), packet-type (4bit 15:12), status (4bit 11:8), 
--            first byte enable (4bit 7:4)
--       Packet type: 0000=wr_req, 0001=rd_request, 0010=rd_completion, 0011=rd_forwarding request, 
--          0100=rd_forw_compl, 0101=wr_forwarding, others: RFU
--       Status: 0000=succesful_transaction, 0001=no_further_hop, 0010=unknown_error
--
-- Addressing: device identification is based on ID. Every packet copies a specified amount
--  of data from source_address in source device to target_address in target device (wr) or 
--  requests a read in the opposite way. The Aurora bridge normally has its own 4G or less address
--  space, which is a DRAM chip connected to the bridge (FPGA), typically 64MBytes. Every board 
--  has its own 0...MAX address space, which is indepenent from the other board's address spaces.
--  The local host is a processor, which can be the board's main x86/PPC/MIPS processor or a dedicated 
--  small processor inside or attached-to the aurora bridge (FPGA).
--  The completion swaps the source/destination ID/address fields!
-- ID: 10-bit chassys address and 6-bit slot address. The slot address can be figured out from 
--  the backplane's geographical addressing pins. The chassys address have to be specified by the 
--  user in a software or stored on the backplane in an EEPROM. The ID is programmed into the 
--  aurora interface registers after powerup by the local host. There is no plug and play device
--  discovery by hardware, although software can initiate discovery by pinging all possible device 
--  numbers in the chassys or in the server room. 
--  ID=0 means that the target of the transaction is the device immediately found at the other end 
--  of the physical link, which also means that chassys addresses start from 1 till 1023, slot 
--  addresses from 0 till 63. ID0 can be used for discovery in backplanes, or for communication
--  in pint-to-point logical/physical connections in dedicated cable-links (system with 2 devices only).
-- Discovery: in the backplane, the port/link connectionsa re known based on the geographical 
--  addressing and the fixed topology. Only slot-0 has access to other chassys. At powerup, every 
--  slot pings the other slots to see if they are there, by a dummy 1-byte read.
-- Multi-hop transactions (FORWARDING) in multi-mesh topologies: if a device receives a packet which does 
--  not match its ID, then it has to store the packet in local DRAM, and the localhost has to retransmit in 
--  the appropriate port (another aurora block), based on the earlier discovery. Packets to other chassys 
--  have to go through slot-0. When a port forwards a packet, the host writes the original source ID into
--  the source ID fifo, not the local ID.
--
--  Registers (aurora_config_wb bus):
--   WRITE REQ (local host writes): source address (local), target address, source ID, target ID, 
--    byte-count.
--   WRITE_COMPLETION (this IP writes it): source address (remote), target address (local), source ID, target ID 
--    (check for match or forwarding), byte-count.
--   READ REQ (local host writes): source address (local), target address, source ID (set up once), target ID, 
--    byte-count.
--   READ completion status: completed_localaddress
--   Forwarding REQUEST received: buffer start address pointer and buffer size (local host writes it, not a FIFO), Packet 
--     pointer (FIFO, the aurora IP writes it). The complete packet with header is stored in memory. the packet size is in the header
--     Forwarding buffer freeup: free up buffer area already forwarded. address and size.
--   Control registers: FIFO status for every FIFO.
--   All of these registers are FIFOs, one transaction involves reading or writing 4 FIFOs. Source ID is local ID 
--    for most of the transactions, except for the forwarded packets. Where more than one FIFO has to be
--    written, write the highest address last, otherwise the scheduler might take a wrong combination.
--    The user must write the local address field last, since that register/FIFO triggers the transfer.
--	  Register Addresses:
--	  -------------------
--		--00h - aurora_fifostatus_reg 
--		--01h - aurora_wr_sourceaddress_reg
--		--02h - aurora_wr_targetaddress_reg
--		--03h - aurora_wr_sourceid_reg 
--		--04h - aurora_wr_targetid_reg
--		--05h - aurora_wr_bytecount_reg
--		--06h - aurora_rdreq_localaddress_reg
--		--07h - aurora_rdreq_targetaddress_reg
--		--08h - aurora_rdreq_localid_reg
--		--09h - aurora_rdreq_targetid_reg
--		--0ah - aurora_rdreq_bytecount_reg
--		--0bh - aurora_rdcompl_localaddress_reg 
--		--0ch - aurora_forwarding_bufferbase_reg
--		--0dh - aurora_forwarding_bufsize_reg 
--		--0eh - aurora_forw_freeupaddr_reg
--		--0fh - aurora_forw_freeupsize_reg
--		--10h - aurora_forwreq_pointer_reg
--    --11h - aurora_arrivedwrite_address_reg
--
--  Interrupts to local host:
--   - read completion (fifo was written)
--   - forwarding request
--   - write arrived
--   if any of the readable FIFOs is not empty, the interrupt stays asserted.
--
----  Trigger for WB transfer:
--   TX: trigger signal from Locan-Host registerbit (aurora_config_wb bus) or request from
--       remote device (response packet as a part of split transaction).
--   RX: arriving packet.
--
-- Refclk: From external pin, on-board 100MHz low jitter clock generator.
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity op2p is
    Port ( --FPGA PINS (EXTERNAL):
           GTPD1_P  : in std_logic; --aurora GTP refclk P
           GTPD1_N  : in std_logic; --aurora GTP refclk N
           aurora_RXP  : in std_logic_vector(0 to 1);
           aurora_RXN  : in std_logic_vector(0 to 1);
           aurora_TXP  : out std_logic_vector(0 to 1);
           aurora_TXN  : out std_logic_vector(0 to 1);
			  --INTERNAL PORTS:
			  --DATA BUS (wishbone): this IP is master on this
			  auroraif_wb_data_o : out std_logic_vector(31 downto 0); 
			  auroraif_wb_data_i : in std_logic_vector(31 downto 0);
			  auroraif_wb_addr_o : out std_logic_vector(25 downto 0);
			  auroraif_wb_cyc_o : out std_logic;
			  auroraif_wb_stb_o : out std_logic;
			  auroraif_wb_wr_o : out std_logic;
			  auroraif_wb_ack_i : in std_logic;
			  auroraif_wb_clk_o : out std_logic; 
			  auroraif_wb_sel_o : out std_logic_vector(3 downto 0);
			  --instruction registers : this IP is slave on this
			  aurora_config_wb_data_o : in std_logic_vector(31 downto 0); 
			  aurora_config_wb_data_i : out std_logic_vector(31 downto 0);
			  aurora_config_wb_addr_i : in std_logic_vector(4 downto 0);
			  aurora_config_wb_cyc_i : in std_logic;
			  aurora_config_wb_stb_i : in std_logic;
			  aurora_config_wb_wr_i : in std_logic;
			  aurora_config_wb_ack_o : out std_logic;
			  aurora_config_wb_clk_i : in std_logic; 	
			  aurora_config_wb_sel_o :  in std_logic_vector(3 downto 0);
			  --system signals			  
			  aurora_reset : in std_logic;
			  x25m_clk : in std_logic;
			  --interrupt out:	
			  aurora_irq : out  STD_LOGIC
			 );
end op2p;

-- ----- ARCHITECTURE START -------------------------------------------------------------------
architecture Behavioral of op2p is

-----------------------------------------------------------------------------------------------
   -- Internal Signals ------------------------------------------------------------------------
	--SIGNAL dummy : std_logic_vector(15 downto 0);	
  SIGNAL  RESET     : std_logic;
  SIGNAL  HARD_ERROR        : std_logic;
  SIGNAL  SOFT_ERROR        : std_logic;
  SIGNAL  FRAME_ERROR       : std_logic;
  SIGNAL  ERR_COUNT         : std_logic_vector(0 to 7);
  SIGNAL  LANE_UP           : std_logic_vector(0 to 1);
  SIGNAL  CHANNEL_UP        : std_logic;
  SIGNAL  INIT_CLK          :  std_logic;
  SIGNAL  GT_RESET_IN       :  std_logic;
  SIGNAL  TX_D    :  std_logic_vector(0 to 31); 
  SIGNAL  TX_REM          :  std_logic_vector(0 to 1);     
  SIGNAL  TX_SOF_N        :  std_logic;
  SIGNAL  TX_EOF_N        :  std_logic;
  SIGNAL  TX_SRC_RDY_N    :  std_logic;
  SIGNAL  TX_DST_RDY_N    :   std_logic;  
  SIGNAL  NFC_REQ_N       :  std_logic;
  SIGNAL  NFC_NB          :  std_logic_vector(0 to 3);     
  SIGNAL  NFC_ACK_N       :   std_logic;  
  SIGNAL  RX_D    :  std_logic_vector(0 to 31); 
  SIGNAL  RX_REM          :  std_logic_vector(0 to 1);     
  SIGNAL  RX_SOF_N        :  std_logic;
  SIGNAL  RX_EOF_N        :  std_logic;
  SIGNAL  RX_SRC_RDY_N    :  std_logic; 
  SIGNAL  RX_SNF          :  std_logic;
  SIGNAL  RX_FC_NB        :  std_logic_vector(0 to 3);  

  --received packet properties -not fifo
  SIGNAL aurora_rx_sourceaddress_reg  : std_logic_vector(31 downto 0);
  SIGNAL aurora_rx_targetaddress_reg : std_logic_vector(31 downto 0);
  SIGNAL aurora_rx_sourceid_reg  : std_logic_vector(31 downto 0);
  SIGNAL aurora_rx_targetid_reg  : std_logic_vector(31 downto 0);
  --SIGNAL aurora_rx_targetid_reg : std_logic_vector(31 downto 0);
  SIGNAL aurora_rx_bytecount_reg  : std_logic_vector(31 downto 0);
  SIGNAL aurora_rx_sourceaddress_regrd_en : std_logic;
  SIGNAL aurora_rx_sourceaddress_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_rx_sourceaddress_regempty : std_logic;
  
  --FIFO signals to the wb-slave interface:
  SIGNAL aurora_wr_sourceaddress_regrd_en : std_logic;
  SIGNAL aurora_wr_targetaddress_regrd_en : std_logic;
  SIGNAL aurora_wr_sourceid_regrd_en : std_logic;
  SIGNAL aurora_wr_targetid_regrd_en : std_logic;
  SIGNAL aurora_wr_bytecount_regrd_en : std_logic;
  SIGNAL aurora_rdreq_localaddress_regrd_en : std_logic;
  SIGNAL aurora_rdreq_targetaddress_regrd_en : std_logic;
  SIGNAL aurora_rdreq_localid_regrd_en : std_logic;
  SIGNAL aurora_rdreq_targetid_regrd_en : std_logic;
  SIGNAL aurora_rdreq_bytecount_regrd_en : std_logic;
  SIGNAL aurora_forw_freeupaddr_regrd_en : std_logic;
  SIGNAL aurora_forw_freeupsize_regrd_en : std_logic;
  SIGNAL aurora_forwreq_pointer_regdin : std_logic_vector(31 downto 0);
  SIGNAL aurora_forwreq_pointer_regwr_en : std_logic;
  SIGNAL aurora_rdcompl_localaddress_regdin : std_logic_vector(31 downto 0);
  SIGNAL aurora_rdcompl_localaddress_regwr_en : std_logic;
  SIGNAL aurora_arrivedwrite_address_regdin : std_logic_vector(31 downto 0);
  SIGNAL aurora_arrivedwrite_address_regwr_en : std_logic;          
  --SIGNAL aurora_config_wb_data_i : std_logic_vector(31 downto 0);
  --SIGNAL aurora_config_wb_ack_o : std_logic;
  --SIGNAL aurora_irq : std_logic;
  SIGNAL aurora_wr_sourceaddress_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_wr_sourceaddress_regempty : std_logic;
  SIGNAL aurora_wr_targetaddress_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_wr_targetaddress_regempty : std_logic;
  SIGNAL aurora_wr_sourceid_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_wr_sourceid_regempty : std_logic;
  SIGNAL aurora_wr_targetid_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_wr_targetid_regempty : std_logic;
  SIGNAL aurora_wr_bytecount_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_wr_bytecount_regempty : std_logic;
  SIGNAL aurora_rdreq_localaddress_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_rdreq_localaddress_regempty : std_logic;
  SIGNAL aurora_rdreq_targetaddress_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_rdreq_targetaddress_regempty : std_logic;
  SIGNAL aurora_rdreq_localid_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_rdreq_localid_regempty : std_logic;
  SIGNAL aurora_rdreq_targetid_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_rdreq_targetid_regempty : std_logic;
  SIGNAL aurora_rdreq_bytecount_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_rdreq_bytecount_regempty : std_logic;
  SIGNAL aurora_forw_freeupaddr_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_forw_freeupaddr_regempty : std_logic;
  SIGNAL aurora_forw_freeupsize_regdout : std_logic_vector(31 downto 0);
  SIGNAL aurora_forw_freeupsize_regempty : std_logic;
  SIGNAL aurora_forwarding_bufferbase_reg : std_logic_vector(31 downto 0);
  SIGNAL aurora_forwarding_bufsize_reg : std_logic_vector(31 downto 0);
  SIGNAL aurora_forwreq_pointer_regfull : std_logic;
  SIGNAL aurora_rdcompl_localaddress_regfull : std_logic;
  SIGNAL aurora_arrivedwrite_address_regfull : std_logic;
  SIGNAL aurora_local_id : std_logic_vector(15 downto 0);

  SIGNAL bram_rxpacket_we  : std_logic_vector(0 downto 0);
  SIGNAL bram_rxpacket_writeaddress : std_logic_vector(8 downto 0);
  SIGNAL bram_rxpacket_writedata : std_logic_vector(31 downto 0);
  SIGNAL bram_rxpacket_readaddress : std_logic_vector(8 downto 0);
  SIGNAL bram_rxpacket_readdata : std_logic_vector(31 downto 0);
  SIGNAL bram_txpacket_we : std_logic_vector(0 downto 0);
  SIGNAL bram_txpacket_writeaddress : std_logic_vector(8 downto 0);
  SIGNAL bram_txpacket_writedata : std_logic_vector(31 downto 0);
  SIGNAL bram_txpacket_readaddress : std_logic_vector(8 downto 0);
  SIGNAL bram_txpacket_readdata : std_logic_vector(31 downto 0);
  
  SIGNAL packetstm_counter : std_logic_vector(7 downto 0);
  SIGNAL fbm_nextraddress_req : std_logic;
  SIGNAL fbm_actualsize  : std_logic_vector(15 downto 0);
  SIGNAL fbm_nextraddress : std_logic_vector(31 downto 0);
  
  SIGNAL target_address : std_logic_vector(31 downto 0);
  SIGNAL source_id : std_logic_vector(15 downto 0);
  SIGNAL target_id : std_logic_vector(15 downto 0);
  SIGNAL source_address : std_logic_vector(31 downto 0);
  SIGNAL byte_count  : std_logic_vector(31 downto 0);

	--
  SIGNAL forw_packetaddr_allocated : std_logic;
  --SIGNAL fbm_nextraddress : std_logic_vector(31 downto 0);
  --SIGNAL fbm_nextraddress_req : std_logic_vector(31 downto 0);

  SIGNAL forwbuf_descriptor_ddword0 : std_logic_vector(63 downto 0);
  SIGNAL forwbuf_descriptor_ddword1 : std_logic_vector(63 downto 0);
  SIGNAL forwbuf_descriptor_ddword2 : std_logic_vector(63 downto 0);
  SIGNAL forwbuf_descriptor_ddword3 : std_logic_vector(63 downto 0);
  SIGNAL forwbuf_descriptor_ddword4 : std_logic_vector(63 downto 0);
  SIGNAL forwbuf_descriptor_ddword5 : std_logic_vector(63 downto 0);
  SIGNAL forwbuf_descriptor_ddword6 : std_logic_vector(63 downto 0);
  SIGNAL forwbuf_descriptor_ddword7 : std_logic_vector(63 downto 0);

  SIGNAL forwbuf_descriptor_byte0 : std_logic_vector(7 downto 0);
  SIGNAL forwbuf_descriptor_byte1 : std_logic_vector(7 downto 0);
  SIGNAL forwbuf_descriptor_byte2 : std_logic_vector(7 downto 0);
  SIGNAL forwbuf_descriptor_byte3 : std_logic_vector(7 downto 0);
  SIGNAL forwbuf_descriptor_byte4 : std_logic_vector(7 downto 0);
  SIGNAL forwbuf_descriptor_byte5 : std_logic_vector(7 downto 0);
  SIGNAL forwbuf_descriptor_byte6 : std_logic_vector(7 downto 0);
  SIGNAL forwbuf_descriptor_byte7 : std_logic_vector(7 downto 0);

  SIGNAL forwbuf_descriptor_bit0 : std_logic;
  SIGNAL forwbuf_descriptor_bit1  : std_logic;
  SIGNAL forwbuf_descriptor_bit2  : std_logic;
  SIGNAL forwbuf_descriptor_bit3 : std_logic;
  SIGNAL forwbuf_descriptor_bit4 : std_logic;
  SIGNAL forwbuf_descriptor_bit5 : std_logic;
  SIGNAL forwbuf_descriptor_bit6 : std_logic;
  SIGNAL forwbuf_descriptor_bit7 : std_logic;
  SIGNAL forwbuff_position_byteoffset : std_logic_vector(18 downto 0);
  SIGNAL forwbuff_position_bitoffset : std_logic_vector(5 downto 0);
  --SIGNAL forwbuff_position_byteoffset : std_logic_vector(18 downto 0);
  SIGNAL forwbuff_position : std_logic_vector(8 downto 0);

  --SIGNAL forwbuff_position_byteoffset : std_logic_vector(18 downto 0);
  --SIGNAL forwbuff_position_bitoffset : std_logic_vector(5 downto 0);
  SIGNAL free_this_location  : std_logic_vector(31 downto 0); --(18 downto 0);
  SIGNAL which_64bit_isit_tofreeup  : std_logic_vector(2 downto 0);
  SIGNAL which_bit_isit_tofreeup : std_logic_vector(5 downto 0);
  SIGNAL fbm_state : std_logic_vector(7 downto 0);
  
  SIGNAL user_clk_i : std_logic;
  
  SIGNAL wb0_state : std_logic_vector(7 downto 0);
  SIGNAL wb_transaction_complete : std_logic;
  SIGNAL auroraif_wb_addr_o_feed : std_logic_vector(31 downto 0);
  SIGNAL auroraif_wb_sel_o_feed : std_logic_vector(3 downto 0);
  SIGNAL auroraif_wb_data_o_feed : std_logic_vector(31 downto 0);
  SIGNAL start_write_wb0 : std_logic;
  SIGNAL auroraif_wb_data_i_latched : std_logic_vector(31 downto 0);
  SIGNAL start_read_wb0 : std_logic;
  SIGNAL rx_fc_nb_latched : std_logic_vector(3 downto 0);
  SIGNAL epif_tx_state : std_logic_vector(7 downto 0);
  SIGNAL pcie_packet_tx_complete : std_logic;
  SIGNAL txtrn_counter : std_logic_vector(7 downto 0);
  SIGNAL packet_state : std_logic_vector(7 downto 0);
  SIGNAL op2p_there_is_a_new_packet_to_transmit : std_logic;
  SIGNAL rxpacket_payloadsize_dwords : std_logic_vector(31 downto 0);
  SIGNAL op2p_just_received_a_new_packet : std_logic;
  SIGNAL epif_rx_state : std_logic_vector(7 downto 0);
  SIGNAL rx_dst_rdy_n : std_logic;
  SIGNAL trn_rx_counter : std_logic_vector(7 downto 0);
  SIGNAL op2p_bar0_wb_data_o_feed : std_logic_vector(31 downto 0);
  SIGNAL op2p_bar0_wb_addr_o_feed : std_logic_vector(31 downto 0);
  SIGNAL op2p_bar0_wb_sel_o_feed : std_logic_vector(3 downto 0);
  SIGNAL rxpacket_decodedaddress : std_logic_vector(31 downto 0);
  SIGNAL rxpacket_firstdw_be : std_logic_vector(3 downto 0);
  SIGNAL rxpacket_lastdw_be : std_logic_vector(3 downto 0);
  SIGNAL rxpacket_requesterid : std_logic_vector(15 downto 0);
  SIGNAL packet_state_copy : std_logic_vector(7 downto 0);
  SIGNAL bram_rxpacket_firstdata_address : std_logic_vector(31 downto 0);
  SIGNAL rxpacket_header_dw1 : std_logic_vector(31 downto 0);
  SIGNAL rxpacket_header_dw2 : std_logic_vector(31 downto 0);
  SIGNAL rxpacket_header_dw3 : std_logic_vector(31 downto 0);
  SIGNAL rxpacket_header_dw4 : std_logic_vector(31 downto 0);
  SIGNAL rxpacket_datacount : std_logic_vector(15 downto 0);
  SIGNAL rxpacket_type : std_logic_vector(3 downto 0);
  SIGNAL rxpacket_status : std_logic_vector(3 downto 0);
  SIGNAL rxpacket_sourceid : std_logic_vector(15 downto 0);
  SIGNAL rxpacket_destinationid : std_logic_vector(15 downto 0);
  SIGNAL rxpacket_targetaddress : std_logic_vector(31 downto 0);
  SIGNAL rxpacket_sourceaddress : std_logic_vector(31 downto 0);
  SIGNAL bit10 : std_logic_vector(1 downto 0);
  SIGNAL op2p_bar0_wb_data_i_latched : std_logic_vector(31 downto 0);
  SIGNAL op2p_packet_tx_complete : std_logic;
  SIGNAL aurora_rdreq_sourceid_regrd_en : std_logic;
  SIGNAL aurora_rdreq_sourceid_regdout : std_logic_vector(31 downto 0);
  SIGNAL trn_reset_n : std_logic;
  SIGNAL cfg_completer_id : std_logic_vector(15 downto 0);
  SIGNAL aurora_forw_freeupaddr_regrd_en_copy : std_logic;
  SIGNAL forwbus_64bit : std_logic_vector(63 downto 0);
  SIGNAL which_64bit_isit : std_logic_vector(2 downto 0);
  SIGNAL forwbus_8bit : std_logic_vector(7 downto 0);
  SIGNAL which_8bit_isit : std_logic_vector(2 downto 0);
  SIGNAL which_bit_isit : std_logic_vector(2 downto 0);
  --SIGNAL 
  --SIGNAL 
  --SIGNAL 
  --SIGNAL 
  --SIGNAL 
  --SIGNAL 
  --SIGNAL 
  --SIGNAL  
 
  
-----------------------------------------------------------------------------------------------
	-- COMPONENT DECLARATIONS (introducing the IPs) --------------------------------------------
	
	component aurora_exmpl_des_modified
	port
	 (
    -- User I/O
            RESET             : in std_logic;
            HARD_ERROR        : out std_logic;
            SOFT_ERROR        : out std_logic;
            FRAME_ERROR       : out std_logic;
            ERR_COUNT         : out std_logic_vector(0 to 7);
            LANE_UP           : out std_logic_vector(0 to 1);
            CHANNEL_UP        : out std_logic;
            INIT_CLK          : in  std_logic;
            GT_RESET_IN       : in  std_logic;
	-- Clocks
           GTPD1_P   : in  std_logic;
           GTPD1_N   : in  std_logic;
			  user_clk_i : out std_logic;
			  x25m_clkin  : in  std_logic;
	--from the Frame Generator in the original reference design: (gtp-tx)
        -- User Interface
        TX_D            : in  std_logic_vector(0 to 31); 
        TX_REM          : in  std_logic_vector(0 to 1);     
        TX_SOF_N        : in  std_logic;
        TX_EOF_N        : in  std_logic;
        TX_SRC_RDY_N    : in  std_logic;
        TX_DST_RDY_N    : out   std_logic;  
        -- NFC Interface
        NFC_REQ_N       : in  std_logic;
        NFC_NB          : in  std_logic_vector(0 to 3);     
        NFC_ACK_N       : out   std_logic;  
	--from the Frame Check module in the original reference design: (gtp-rx)
        -- User Interface
        RX_D            : out  std_logic_vector(0 to 31); 
        RX_REM          : out  std_logic_vector(0 to 1);     
        RX_SOF_N        : out  std_logic;
        RX_EOF_N        : out  std_logic;
        RX_SRC_RDY_N    : out  std_logic; 
        -- NFC Interface
        RX_SNF          : out  std_logic;
        RX_FC_NB        : out  std_logic_vector(0 to 3);   
   -- Gbit I/O
            RXP               : in std_logic_vector(0 to 1);
            RXN               : in std_logic_vector(0 to 1);
            TXP               : out std_logic_vector(0 to 1);
            TXN               : out std_logic_vector(0 to 1)
	 );
	end component;

	COMPONENT op2p_host_wb_if
	PORT(
		aurora_config_wb_data_o : IN std_logic_vector(31 downto 0);
		aurora_config_wb_addr_i : IN std_logic_vector(4 downto 0);
		aurora_config_wb_cyc_i : IN std_logic;
		aurora_config_wb_stb_i : IN std_logic;
		aurora_config_wb_wr_i : IN std_logic;
		aurora_config_wb_clk_i : IN std_logic;
		aurora_config_wb_sel_o : IN std_logic_vector(3 downto 0);
		aurora_reset : IN std_logic;
		x25m_clk : IN std_logic;
		user_clk_i : in std_logic;
		aurora_wr_sourceaddress_regrd_en : IN std_logic;
		aurora_wr_targetaddress_regrd_en : IN std_logic;
		aurora_wr_sourceid_regrd_en : IN std_logic;
		aurora_wr_targetid_regrd_en : IN std_logic;
		aurora_wr_bytecount_regrd_en : IN std_logic;
		aurora_rdreq_localaddress_regrd_en : IN std_logic;
		aurora_rdreq_targetaddress_regrd_en : IN std_logic;
		aurora_rdreq_localid_regrd_en : IN std_logic;
		aurora_rdreq_targetid_regrd_en : IN std_logic;
		aurora_rdreq_bytecount_regrd_en : IN std_logic;
		aurora_forw_freeupaddr_regrd_en : IN std_logic;
		aurora_forw_freeupsize_regrd_en : IN std_logic;
		aurora_forwreq_pointer_regdin : IN std_logic_vector(31 downto 0);
		aurora_forwreq_pointer_regwr_en : IN std_logic;
		aurora_rdcompl_localaddress_regdin : IN std_logic_vector(31 downto 0);
		aurora_rdcompl_localaddress_regwr_en : IN std_logic;
		aurora_arrivedwrite_address_regdin : IN std_logic_vector(31 downto 0);
		aurora_arrivedwrite_address_regwr_en : IN std_logic;          
		aurora_config_wb_data_i : OUT std_logic_vector(31 downto 0);
		aurora_config_wb_ack_o : OUT std_logic;
		aurora_irq : OUT std_logic;
		aurora_wr_sourceaddress_regdout : OUT std_logic_vector(31 downto 0);
		aurora_wr_sourceaddress_regempty : OUT std_logic;
		aurora_wr_targetaddress_regdout : OUT std_logic_vector(31 downto 0);
		aurora_wr_targetaddress_regempty : OUT std_logic;
		aurora_wr_sourceid_regdout : OUT std_logic_vector(31 downto 0);
		aurora_wr_sourceid_regempty : OUT std_logic;
		aurora_wr_targetid_regdout : OUT std_logic_vector(31 downto 0);
		aurora_wr_targetid_regempty : OUT std_logic;
		aurora_wr_bytecount_regdout : OUT std_logic_vector(31 downto 0);
		aurora_wr_bytecount_regempty : OUT std_logic;
		aurora_rdreq_localaddress_regdout : OUT std_logic_vector(31 downto 0);
		aurora_rdreq_localaddress_regempty : OUT std_logic;
		aurora_rdreq_targetaddress_regdout : OUT std_logic_vector(31 downto 0);
		aurora_rdreq_targetaddress_regempty : OUT std_logic;
		aurora_rdreq_localid_regdout : OUT std_logic_vector(31 downto 0);
		aurora_rdreq_localid_regempty : OUT std_logic;
		aurora_rdreq_targetid_regdout : OUT std_logic_vector(31 downto 0);
		aurora_rdreq_targetid_regempty : OUT std_logic;
		aurora_rdreq_bytecount_regdout : OUT std_logic_vector(31 downto 0);
		aurora_rdreq_bytecount_regempty : OUT std_logic;
		aurora_local_id : out  std_logic_vector(15 downto 0); 
		aurora_forw_freeupaddr_regdout : OUT std_logic_vector(31 downto 0);
		aurora_forw_freeupaddr_regempty : OUT std_logic;
		aurora_forw_freeupsize_regdout : OUT std_logic_vector(31 downto 0);
		aurora_forw_freeupsize_regempty : OUT std_logic;
		aurora_forwarding_bufferbase_reg : OUT std_logic_vector(31 downto 0);
		aurora_forwarding_bufsize_reg : OUT std_logic_vector(31 downto 0);
		aurora_forwreq_pointer_regfull : OUT std_logic;
		aurora_rdcompl_localaddress_regfull : OUT std_logic;
		aurora_arrivedwrite_address_regfull : OUT std_logic
		);
	END COMPONENT;

	COMPONENT blk_mem_gen_v4_1
	PORT(
		clka : IN std_logic;
		wea : IN std_logic_vector(0 to 0);
		addra : IN std_logic_vector(8 downto 0);
		dina : IN std_logic_vector(31 downto 0);
		clkb : IN std_logic;
		addrb : IN std_logic_vector(8 downto 0);          
		doutb : OUT std_logic_vector(31 downto 0)
		);
	END COMPONENT;


-----------------------------------------------------------------------------------------------
-- --------ARCHITECTURE BODY BEGINS -----------------------------------------------------------
begin



-----------------------------------------------------------------------------------------------
	-- COMPONENT INSTALLATIONS (connecting the IPs to local signals) ---------------------------
	Inst_aurora_exmpl_des_modified : aurora_exmpl_des_modified
	  port map
		(
		--  User I/O      
		 RESET   =>  aurora_reset  ,   -- : in std_logic;     
		 HARD_ERROR   =>  HARD_ERROR  ,   -- : out std_logic;     
		 SOFT_ERROR   =>  SOFT_ERROR  ,   -- : out std_logic;     
		 FRAME_ERROR   =>  FRAME_ERROR  ,   -- : out std_logic;     
		 ERR_COUNT   =>  ERR_COUNT  ,   -- : out std_logic_vector(0 to 7);   
		 LANE_UP   =>  LANE_UP  ,   -- : out std_logic_vector(0 to 1);   
		 CHANNEL_UP   =>  CHANNEL_UP  ,   -- : out std_logic;     
		 INIT_CLK   =>  INIT_CLK  ,   -- : in std_logic;     
		 GT_RESET_IN   =>  GT_RESET_IN  ,   -- : in std_logic;     
		 -- Clocks       
		 GTPD1_P   =>  GTPD1_P  ,   -- : in std_logic;     
		 GTPD1_N   =>  GTPD1_N  ,   -- : in std_logic;
		 user_clk_i => user_clk_i,
		 x25m_clkin  => x25m_clk,
		-- the Frame Generator in the original reference design:
		 -- User Interface      
		 TX_D   =>  TX_D  ,   -- : out std_logic_vector(0 to 31);   
		 TX_REM   =>  TX_REM  ,   -- : out std_logic_vector(0 to 1);    
		 TX_SOF_N   =>  TX_SOF_N  ,   -- : out std_logic;      
		 TX_EOF_N   =>  TX_EOF_N  ,   -- : out std_logic;      
		 TX_SRC_RDY_N   =>  TX_SRC_RDY_N  ,   -- : out std_logic;      
		 TX_DST_RDY_N   =>  TX_DST_RDY_N  ,   -- : in std_logic;      
		 -- NFC Interface       
		 NFC_REQ_N   =>  NFC_REQ_N  ,   -- : out std_logic;      
		 NFC_NB   =>  NFC_NB  ,   -- : out std_logic_vector(0 to 3);    
		 NFC_ACK_N   =>  NFC_ACK_N  ,   -- : in std_logic;        
		-- the Frame Check module in the original reference design:
		 -- User Interface       
		 RX_D   =>  RX_D  ,   -- : in std_logic_vector(0 to 31);    
		 RX_REM   =>  RX_REM  ,   -- : in std_logic_vector(0 to 1);
		 RX_SOF_N   =>  RX_SOF_N  ,   -- : in std_logic;  
		 RX_EOF_N   =>  RX_EOF_N  ,   -- : in std_logic;  
		 RX_SRC_RDY_N   =>  RX_SRC_RDY_N  ,   -- : in std_logic;  
		 -- NFC Interface   
		 RX_SNF   =>  RX_SNF  ,   -- : in std_logic;  
		 RX_FC_NB   =>  RX_FC_NB  ,   -- : in std_logic_vector(0 to 3); 
		-- V5 I/O   
		 RXP   =>  aurora_RXP  ,   -- : in std_logic_vector(0 to 1);
		 RXN   =>  aurora_RXN  ,   -- : in std_logic_vector(0 to 1);
		 TXP   =>  aurora_TXP  ,   -- : out std_logic_vector(0 to 1);
		 TXN   =>  aurora_TXN     -- : out std_logic_vector(0 to 1)
		 ); 


	--the wishbone slave interface with its FIFOs.
	Inst_op2p_host_wb_if: op2p_host_wb_if PORT MAP(
	 aurora_config_wb_data_o => aurora_config_wb_data_o ,
	 aurora_config_wb_data_i => aurora_config_wb_data_i ,
	 aurora_config_wb_addr_i => aurora_config_wb_addr_i ,
	 aurora_config_wb_cyc_i => aurora_config_wb_cyc_i ,
	 aurora_config_wb_stb_i => aurora_config_wb_stb_i ,
	 aurora_config_wb_wr_i => aurora_config_wb_wr_i ,
	 aurora_config_wb_ack_o => aurora_config_wb_ack_o ,
	 aurora_config_wb_clk_i => aurora_config_wb_clk_i ,
	 aurora_config_wb_sel_o => aurora_config_wb_sel_o ,
	 aurora_reset => aurora_reset ,
	 x25m_clk => x25m_clk ,
	 user_clk_i => user_clk_i,
	 aurora_irq => aurora_irq ,
	 aurora_wr_sourceaddress_regrd_en => aurora_wr_sourceaddress_regrd_en ,
	 aurora_wr_sourceaddress_regdout => aurora_wr_sourceaddress_regdout ,
	 aurora_wr_sourceaddress_regempty => aurora_wr_sourceaddress_regempty ,
	 aurora_wr_targetaddress_regrd_en => aurora_wr_targetaddress_regrd_en ,
	 aurora_wr_targetaddress_regdout => aurora_wr_targetaddress_regdout ,
	 aurora_wr_targetaddress_regempty => aurora_wr_targetaddress_regempty ,
	 aurora_wr_sourceid_regrd_en => aurora_wr_sourceid_regrd_en ,
	 aurora_wr_sourceid_regdout => aurora_wr_sourceid_regdout ,
	 aurora_wr_sourceid_regempty => aurora_wr_sourceid_regempty ,
	 aurora_wr_targetid_regrd_en => aurora_wr_targetid_regrd_en ,
	 aurora_wr_targetid_regdout => aurora_wr_targetid_regdout ,
	 aurora_wr_targetid_regempty => aurora_wr_targetid_regempty ,
	 aurora_wr_bytecount_regrd_en => aurora_wr_bytecount_regrd_en ,
	 aurora_wr_bytecount_regdout => aurora_wr_bytecount_regdout ,
	 aurora_wr_bytecount_regempty => aurora_wr_bytecount_regempty ,
	 aurora_rdreq_localaddress_regrd_en => aurora_rdreq_localaddress_regrd_en ,
	 aurora_rdreq_localaddress_regdout => aurora_rdreq_localaddress_regdout ,
	 aurora_rdreq_localaddress_regempty => aurora_rdreq_localaddress_regempty ,
	 aurora_rdreq_targetaddress_regrd_en => aurora_rdreq_targetaddress_regrd_en ,
	 aurora_rdreq_targetaddress_regdout => aurora_rdreq_targetaddress_regdout ,
	 aurora_rdreq_targetaddress_regempty => aurora_rdreq_targetaddress_regempty ,
	 aurora_rdreq_localid_regrd_en => aurora_rdreq_localid_regrd_en ,
	 aurora_rdreq_localid_regdout => aurora_rdreq_localid_regdout ,
	 aurora_rdreq_localid_regempty => aurora_rdreq_localid_regempty ,
	 aurora_rdreq_targetid_regrd_en => aurora_rdreq_targetid_regrd_en ,
	 aurora_rdreq_targetid_regdout => aurora_rdreq_targetid_regdout ,
	 aurora_rdreq_targetid_regempty => aurora_rdreq_targetid_regempty ,
	 aurora_rdreq_bytecount_regrd_en => aurora_rdreq_bytecount_regrd_en ,
	 aurora_rdreq_bytecount_regdout => aurora_rdreq_bytecount_regdout ,
	 aurora_rdreq_bytecount_regempty => aurora_rdreq_bytecount_regempty ,
	 aurora_local_id => aurora_local_id,
	 aurora_forw_freeupaddr_regrd_en => aurora_forw_freeupaddr_regrd_en ,
	 aurora_forw_freeupaddr_regdout => aurora_forw_freeupaddr_regdout ,
	 aurora_forw_freeupaddr_regempty => aurora_forw_freeupaddr_regempty ,
	 aurora_forw_freeupsize_regrd_en => aurora_forw_freeupsize_regrd_en ,
	 aurora_forw_freeupsize_regdout => aurora_forw_freeupsize_regdout ,
	 aurora_forw_freeupsize_regempty => aurora_forw_freeupsize_regempty ,
	 aurora_forwarding_bufferbase_reg => aurora_forwarding_bufferbase_reg ,
	 aurora_forwarding_bufsize_reg => aurora_forwarding_bufsize_reg ,
	 aurora_forwreq_pointer_regdin => aurora_forwreq_pointer_regdin ,
	 aurora_forwreq_pointer_regwr_en => aurora_forwreq_pointer_regwr_en ,
	 aurora_forwreq_pointer_regfull => aurora_forwreq_pointer_regfull ,
	 aurora_rdcompl_localaddress_regdin => aurora_rdcompl_localaddress_regdin ,
	 aurora_rdcompl_localaddress_regwr_en => aurora_rdcompl_localaddress_regwr_en ,
	 aurora_rdcompl_localaddress_regfull => aurora_rdcompl_localaddress_regfull ,
	 aurora_arrivedwrite_address_regdin => aurora_arrivedwrite_address_regdin ,
	 aurora_arrivedwrite_address_regwr_en => aurora_arrivedwrite_address_regwr_en ,
	 aurora_arrivedwrite_address_regfull => aurora_arrivedwrite_address_regfull 
	);

	--block ram for RX packet:
	Inst_bram_rxpacket: blk_mem_gen_v4_1 PORT MAP(
		clka => user_clk_i,
		wea => bram_rxpacket_we,
		addra => bram_rxpacket_writeaddress(8 downto 0),
		dina => bram_rxpacket_writedata,
		clkb => user_clk_i,
		addrb => bram_rxpacket_readaddress(8 downto 0),
		doutb => bram_rxpacket_readdata
	);

	--block ram for TX packet:
	Inst_bram_txpacket: blk_mem_gen_v4_1 PORT MAP(
		clka => user_clk_i,
		wea => bram_txpacket_we,
		addra => bram_txpacket_writeaddress(8 downto 0),
		dina => bram_txpacket_writedata,
		clkb => user_clk_i,
		addrb => bram_txpacket_readaddress(8 downto 0),
		doutb => bram_txpacket_readdata
	); 
				
				
-----------------------------------------------------------------------------------------------
	-- MAIN LOGIC: -----------------------------------------------------------------------------


	--aurora interface FIX conenctions:
	RESET <= aurora_reset;
	trn_reset_n <= not aurora_reset;
	INIT_CLK <= x25m_clk; --Debounce the GT_RESET_IN signal using the INIT_CLK
	GT_RESET_IN <= aurora_reset;
	auroraif_wb_clk_o <=  user_clk_i;
	




	-- -----------------------------------------------------------------------------
	-- WISBONE Master INTERFACE ----------------------------------------------------
	--This is for accessing the local memory.

    --main state machine: set states, capture inputs, set addr/data outputs
	 --minimum 2 clock cycles / transaction. writes are posted, reads have wait states.
    process (aurora_reset, user_clk_i, wb0_state, start_read_wb0, start_write_wb0,
				auroraif_wb_addr_o_feed, auroraif_wb_data_o_feed, auroraif_wb_sel_o_feed) 
    begin
    if (aurora_reset='0') then 
       wb0_state <= "00000000";
       wb_transaction_complete <= '0';
		 auroraif_wb_addr_o <= "00000000000000000000000000";
		 auroraif_wb_sel_o <= "0000";
		 auroraif_wb_data_o <= "00000000000000000000000000000000";			 
    else
      if (user_clk_i'event and user_clk_i = '1') then
                case ( wb0_state ) is

                --********** IDLE STATE  **********
                when "00000000" =>   --state 0        
                    wb_transaction_complete <='0';
						  if (start_read_wb0 ='1') then --go to read
						    wb0_state <= "00000001";
							 auroraif_wb_addr_o <= auroraif_wb_addr_o_feed(25 downto 0);
							 auroraif_wb_sel_o <= auroraif_wb_sel_o_feed;
						  elsif (start_write_wb0 ='1') then --go to write
						    wb0_state <= "00000010";
							 auroraif_wb_addr_o <= auroraif_wb_addr_o_feed (25 downto 0);
							 auroraif_wb_sel_o <= auroraif_wb_sel_o_feed;
							 auroraif_wb_data_o <= auroraif_wb_data_o_feed;
						  end if;

                --********** READ STATE ********** 
					 --set the outputs, 
					 --if ACK asserted, sample the data input
					 --The hold requirements are oversatisfyed by going back to idle, 
					 --and by the fact that the slave uses the cyc/stb/wr strobes synchronously.
                when "00000001" =>   --state 1
                    if (auroraif_wb_ack_i='1') then
						    auroraif_wb_data_i_latched <= auroraif_wb_data_i; --sample the incoming data
							 wb_transaction_complete <='1'; --signalling ready, but only for one clock cycle
							 wb0_state <= "00000000"; --go to state 0
						  else
						  	 wb_transaction_complete <='0';
						  end if;	   		  

                --********** WRITE STATE **********     
					 --if ACK asserted, go back to idle
					 --The hold requirements are oversatisfyed by waiting for ACK to remove write data					 
                when "00000010" =>   --state 2
                    if (auroraif_wb_ack_i='1') then
							 wb0_state <= "00000000"; --go to state 0
							 wb_transaction_complete <='1';
						  else
						     wb_transaction_complete <='0';
						  end if;
						  
                when others => --error
                      wb0_state <= "00000000"; --go to state 0
                end case;     
       end if;        
    end if;
    end process;
    --sync control on wb-control signals:
    process (aurora_reset, wb0_state) 
    begin
    if (aurora_reset='0') then 
		auroraif_wb_cyc_o  <= '0';
		auroraif_wb_stb_o  <= '0';
		auroraif_wb_wr_o  <= '0';
    else
      if (wb0_state = 0) then --idle
			auroraif_wb_cyc_o  <= '0';
			auroraif_wb_stb_o  <= '0';
			auroraif_wb_wr_o  <= '0';
      elsif (wb0_state = 1) then --read 
			auroraif_wb_cyc_o  <= '1';
			auroraif_wb_stb_o  <= '1';
			auroraif_wb_wr_o  <= '0';
      elsif (wb0_state = 10) then --write 
			auroraif_wb_cyc_o  <= '1';
			auroraif_wb_stb_o  <= '1';
			auroraif_wb_wr_o  <= '1';
		else
			auroraif_wb_cyc_o  <= '0';
			auroraif_wb_stb_o  <= '0';
			auroraif_wb_wr_o  <= '0';
		end if;
    end if;
    end process;







	-- -----------------------------------------------------------------------------------
	-- INTERFACE TO THE aurora-IP --------------------------------------------------------
	
	--what to do with these?
        --TX_REM          : in  std_logic_vector(0 to 1); 
		  --Specifies the number of valid bytes in the last data beat;
		  --valid only while TX_EOF_N is asserted. REM bus widths
		  --are given by [0:r(n)], where r(n) = ceiling [{log2(n)}-1].
		  --The TX_REM bus is used to indicate the number of valid bytes in the final word of the frame.
		  TX_REM <= "11"; --always all 4 bytes of a dword valid.
		  
        --RX_REM          : out  std_logic_vector(0 to 1); 
		  --Specifies the number of valid bytes in the last data beat;
		  --valid only when RX_EOF_N is asserted. REM bus widths
		  --are given by [0:r(n)], where r(n) = ceiling [{log2(n)}-1].
		  --dont conenct.

	--FLOW CONTROL ---------------------------------------
	--Flow control logic is built into the TX statemachine.
	--The Aurora 8B/10B protocol includes native flow control (NFC) to allow receivers to
	--control the rate at which data is sent to them by specifying a number of idle data beats that
	--must be placed into the data stream. The data flow can even be turned off completely by
	--requesting that the transmitter temporarily send only idles (XOFF).
	--To send an NFC message to a channel partner, the user application asserts NFC_REQ_N
	--and writes an NFC code to NFC_NB. The NFC code indicates the minimum number of idle
	--cycles the channel partner should insert in its TX data stream. The user application must
	--hold NFC_REQ_N and NFC_NB until NFC_ACK_N is asserted on a positive USER_CLK
	--edge, indicating the Aurora 8B/10B core will transmit the NFC message.
	--Aurora 8B/10B cores cannot transmit data while sending NFC messages.
	--
	 --In this code, we only support NFC_NB=1111=wait or 0000=ready
	 --only allow packets to arrive when the packet buffer is not in use by the scheduler. 
	 
	 --latch the latest flow control state. 0000=ready, 1111=wait, others not supported (assumes wait)
	 process ( aurora_reset, user_clk_i, RX_FC_NB)
    begin
       if (aurora_reset='1') then
         RX_FC_NB_latched <= RX_FC_NB;
       elsif (user_clk_i'event and user_clk_i='1') then
         if (RX_SNF='1') then
			  RX_FC_NB_latched <= RX_FC_NB;
			end if;
       end if;
    end process;	


	-- TX: INTERFACE TO THE aurora-IP: TRANSMIT  PACKETS:-----
    process (aurora_reset, user_clk_i, epif_tx_state, bram_txpacket_readdata, bram_txpacket_readaddress,
				op2p_there_is_a_new_packet_to_transmit, rxpacket_payloadsize_dwords, txtrn_counter, 
				NFC_ACK_N, RX_SNF, RX_FC_NB, aurora_forwreq_pointer_regfull) 
    begin
    if (aurora_reset='0') then 
      epif_tx_state <= "00000000";
      TX_SRC_RDY_N <='1';
		TX_SOF_N <= '1';
		TX_EOF_N <= '1';
		TX_D <= (OTHERS => '0');
		pcie_packet_tx_complete <= '0';
		txtrn_counter <= "00000001";
		bram_txpacket_readaddress <= (OTHERS => '0');
		NFC_REQ_N <= '1'; --flow control wait request off
    else
      if (user_clk_i'event and user_clk_i = '1') then
                case ( epif_tx_state ) is

                --********** idle STATE  **********
                when "00000000" =>   --state 0        
                    --if there is a new packet assembled and the EP is ready, 
						  --start the tx-trn bus transaction.
						  if (op2p_there_is_a_new_packet_to_transmit='1' --scheduler ordered a transmit
								and TX_DST_RDY_N='0'							--aurora ip ready
								and RX_FC_NB_latched = "0000") then    --other end did not request a wait
						    epif_tx_state <= "00000001"; --next state
						  elsif (aurora_forwreq_pointer_regfull='1' --dont send more forwarding or any packet
									or packet_state="00000001" --the local scheduler is processing the last received packed data
									or packet_state="00000010" 
									or packet_state="00000110" 
									or packet_state="00000111" 
									or packet_state="00001000") then --scheduler's read state 
						    epif_tx_state <= "00000101"; --wait state
						  end if;
                    TX_SRC_RDY_N <='1';
						  TX_SOF_N <= '1';
						  TX_EOF_N <= '1';
						  TX_D <= (OTHERS => '0');
						  pcie_packet_tx_complete <= '0';
						  txtrn_counter <= "00000001";
						  bram_txpacket_readaddress <= (OTHERS => '0');
						  NFC_REQ_N <= '1'; --flow control wait request off

                --********** first byte of header STATE **********
                when "00000001" =>   --state 1
                    TX_SRC_RDY_N <='0';
						  TX_SOF_N <= '0'; --start
						  TX_EOF_N <= '1';
						  TX_D <= bram_txpacket_readdata;
						  pcie_packet_tx_complete <= '0';
						  if (TX_DST_RDY_N='0') then
						    txtrn_counter <= txtrn_counter  +1;
						    bram_txpacket_readaddress <= bram_txpacket_readaddress +1;
						  end if;
						  epif_tx_state <= "00000010"; --next
						  
                --********** consecutive dwords STATE **********     				 
                when "00000011" =>   --state 3
                    TX_SRC_RDY_N <='0';
						  TX_SOF_N <= '1';
						  if (TX_DST_RDY_N='0') then
						    txtrn_counter <= txtrn_counter  +1;
						    bram_txpacket_readaddress <= bram_txpacket_readaddress +1;
						  end if;
						  TX_D <=  bram_txpacket_readdata;
						  --test number of dwords:
						  if (txtrn_counter = rxpacket_payloadsize_dwords) then
						  --this is the last dword, next clk is next state
							 epif_tx_state <= "00000000"; --back to idle, since finished
						    TX_EOF_N <= '0'; --end
						    pcie_packet_tx_complete <= '1'; --assert for 1 clk
						  else
						    TX_EOF_N <= '1'; --not end yet
						    pcie_packet_tx_complete <= '0'; --not complete yet
						  end if;

                --********** flow control wait request state STATE **********
					 --ask the link partner to wait
                when "00000100" =>   --state 1
                    NFC_NB <= "1111"; --pause: do not send me more data for now
						  if ( NFC_ACK_N='0') then
						    NFC_REQ_N <= '1'; --flow control wait request off
							 epif_tx_state <= "00000101"; --back to idle, but only after reenabling
 						  else
						    NFC_REQ_N <= '0'; --flow control wait request on
						  end if;

                --********** flow control re-enable STATE ********** 
					 --Re-enable (for the link-partner) the transmission , if the scheduler 
					 --is in idle, or at least not currently using the RX_packet packet buffer.
                when "00000101" =>   --state 1
                    NFC_NB <= "0000"; --enable data flow
						  if (packet_state="00000000" and aurora_forwreq_pointer_regfull='0') then
						  --or use this:  (packet_state/="xxxxxxxx") then
							  if ( NFC_ACK_N='0') then
								 NFC_REQ_N <= '1'; --flow control wait request off
								 epif_tx_state <= "00000000"; --back to idle
							  else
								 NFC_REQ_N <= '0'; --flow control wait request on
							  end if;
						  else
						    NFC_REQ_N <= '1'; --flow control wait request off
						  end if;

                when others => --error
                    epif_tx_state <= "00000000"; --back to idle
                    TX_SRC_RDY_N <='1';
						  TX_SOF_N <= '1';
						  TX_EOF_N <= '1';
						  TX_D <= (OTHERS => '0');
						  pcie_packet_tx_complete <= '0';
						  txtrn_counter <= "00000001";
						  
                end case;     
       end if;        
    end if;
    end process;
	 



	-- RX: INTERFACE TO THE aurora-IP: GET the received packet PACKETS:- ----
    process (aurora_reset, user_clk_i, epif_rx_state, packet_state, trn_rx_counter,
				bram_rxpacket_writeaddress) 
    begin
    if (aurora_reset='0') then 
       bram_rxpacket_writedata <= (OTHERS => '0');
		 bram_rxpacket_we <=  "0"; 
		 op2p_just_received_a_new_packet <= '0'; 
		 epif_rx_state	<= "00000000"; 
		 RX_Dst_rdy_n <= '1';
		 trn_rx_counter <= (OTHERS => '0');
		 bram_rxpacket_writeaddress <=  (OTHERS => '0');
    else
      if (user_clk_i'event and user_clk_i = '1') then

            if (packet_state = 0)then
				  RX_Dst_rdy_n <= '0';
				else
				  RX_Dst_rdy_n <= '1';
				end if;		
		
                case ( epif_rx_state ) is

                --********** idle STATE  **********
                when "00000000" =>   --state 0
						  op2p_just_received_a_new_packet <= '0';
						  if (RX_SRC_RDY_N='0' and RX_SOF_N ='0') then 
						    trn_rx_counter <= trn_rx_counter +1;
							 epif_rx_state <= "00000001";
							 --read first DW: 
							 bram_rxpacket_writedata <= RX_D;
							 bram_rxpacket_writeaddress <= bram_rxpacket_writeaddress +1;
							 bram_rxpacket_we <=  "1"; 
						  else
						    trn_rx_counter <= (OTHERS => '0');
							 bram_rxpacket_we <=  "0"; 
							 bram_rxpacket_writeaddress <=  (OTHERS => '0');
						  end if;

                --********** read STATE ********** 
                when "00000001" =>   --state 1
						  if (RX_EOF_N ='0') then --last dw
						    epif_rx_state <= "00000010"; --for the next clk cycle
						  end if;
						  if (RX_SRC_RDY_N='0') then --only act if the EP was ready
							  trn_rx_counter <= trn_rx_counter +1;
							  --get the data:
							  bram_rxpacket_writedata <= RX_D;
							  bram_rxpacket_we <=  "1"; 
							  bram_rxpacket_writeaddress <= bram_rxpacket_writeaddress +1;
						  end if;

                --********** finished filling up RX packet STATE **********     				 
                when "00000010" =>   --state 2
                    op2p_just_received_a_new_packet <= '1';--assert for one clk only
						  epif_rx_state <= "00000000";
						  trn_rx_counter <= (OTHERS => '0');
						  bram_rxpacket_we <=  "0";
						  bram_rxpacket_writeaddress <=  (OTHERS => '0');
						  
                when others => --error
                      epif_rx_state <= "00000000"; --go to state 0
                end case;     
       end if;        
    end if;
    end process;
	 






	-- -----------------------------------------------------------------------------------
	-- --- MAIN SCHEDULER ----------------------------------------------------------------
	-- --- ALSO PACKET PROCESSING.

	-- Packet format: Header + payload data. Max 2kBytes in one packet. If the request is tergating
	--  a larger than 2kBytes memory area, then the scheduler has to divide the data into 2kB packets.
	--  Header: 
	--    1st DW: source ID (16bit 31:16), destination ID (16bit 15:0)
	--    2nd DW: target address (32bit)
	--    3rd DW: source address (32bit)
	--    4th DW: byte count(16bit: 1-64k 31:16), packet-type (4bit 15:12), status (4bit 11:8), 
	--            first byte enable (4bit 7:4)
	--       Packet type: 0000=wr_req, 0001=rd_request, 0010=rd_completion, 0011=rd_forwarding request, 
	--          0100=rd_forw_compl, 0101=wr_forwarding, others: RFU
	--       Status: 0000=succesful_transaction, 0001=no_further_hop, 0010=unknown_error
	-- The user must write the local address field last, since that register/FIFO triggers the transfer.


	 --packet-protocol statemachine:
    process (trn_reset_n, user_clk_i, packet_state, 
				op2p_just_received_a_new_packet, rxpacket_datacount, aurora_local_id,
				bram_rxpacket_readdata,  bram_txpacket_writeaddress, bram_rxpacket_readaddress,
				packet_state_copy,  rxpacket_decodedaddress, bram_rxpacket_firstdata_address,
				rxpacket_header_dw1, rxpacket_header_dw2, rxpacket_header_dw3, rxpacket_header_dw4,
				bit10, rxpacket_firstdw_be, wb_transaction_complete,
				rxpacket_payloadsize_dwords, op2p_bar0_wb_data_i_latched, cfg_completer_id,
				rxpacket_requesterid) 
    begin
    if (trn_reset_n='0') then 
		start_read_wb0 <= '0';
		start_write_wb0 <= '0';
		op2p_bar0_wb_data_o_feed	 <= (others => '0');
		op2p_bar0_wb_addr_o_feed <= (others => '0');
		op2p_bar0_wb_sel_o_feed  <= (others => '0');
		op2p_there_is_a_new_packet_to_transmit  <= '0';
		rxpacket_decodedaddress<= (others => '0');
		rxpacket_payloadsize_dwords <= (others => '0');
		rxpacket_firstdw_be <= (others => '0');
		rxpacket_lastdw_be <= (others => '0');
		rxpacket_requesterid <= (others => '0');
		packet_state <= (others => '0');
		packet_state_copy  <= (others => '0');
		bram_txpacket_we <= "0";
		bram_txpacket_writeaddress    <= (others => '0');
		bram_txpacket_writedata     <= (others => '0');
		bram_rxpacket_readaddress   <= (others => '0');
		bram_rxpacket_firstdata_address   <= (others => '0');
		rxpacket_header_dw1   <= (others => '0');
		rxpacket_header_dw2   <= (others => '0');
		rxpacket_header_dw3   <= (others => '0');
		rxpacket_header_dw4   <= "00000000000000001111000000000000";
		packetstm_counter <=  (others => '0');
		fbm_nextraddress_req <= '0';
		target_address <=  (others => '0');
		source_id <=  (others => '0');
		target_id <=  (others => '0');
		source_address <=  (others => '0');
		byte_count <=  (others => '0');
    else
      if (user_clk_i'event and user_clk_i = '1') then
                case ( packet_state ) is

                --********** IDLE STATE  **********
                when "00000000" =>   --state 0   
						  --request from link partner: (write, read req, forw req, forw-compl)					 
                    if (op2p_just_received_a_new_packet='1') then
						    packet_state <= "00000001"; --to packet decoding state
                    --request from local host: (write req, read req, forw compl, forw req)
						  elsif (aurora_wr_sourceaddress_regempty = '0') then
						    packet_state <= "00010000"; --to outgoing write state
						  elsif (aurora_rdreq_localaddress_regempty = '0') then
						    packet_state <= "01000000"; --to outgoing RDREQ state
						  end if;
						  --general preparation:				
						  start_write_wb0 <= '0';
						  start_read_wb0 <= '0';
						  packet_state_copy <= packet_state;
							bram_txpacket_we <= "0";
							bram_txpacket_writeaddress   <= (others => '0');
							bram_txpacket_writedata     <= (others => '0');
							bram_rxpacket_readaddress    <= (others => '0');
							rxpacket_datacount <= (others => '0');
							bram_rxpacket_firstdata_address   <= (others => '0');
							packetstm_counter <=  (others => '0');
							rxpacket_header_dw4   <= "00000000000000001111000000000000"; --this is to prevent false decode
				
						  
                --********** packet ARRIVED STATE **********
					 --initiated by link partner
					 --read packet out of EP, decode and decide,
					 --latch address/sel/wr_data
                when "00000001" =>   --state 1
						  if (bram_rxpacket_readaddress < "000000011") then
						    bram_rxpacket_readaddress <= bram_rxpacket_readaddress;
						  else
						    bram_rxpacket_readaddress <= bram_rxpacket_readaddress + 1;
						  end if;
						  packetstm_counter <=  packetstm_counter + 1;
						  --latch the header:
						  if (bram_rxpacket_readaddress = "000000000") then
						    rxpacket_header_dw1 <= bram_rxpacket_readdata;
						  elsif (bram_rxpacket_readaddress = "000000001") then
						    rxpacket_header_dw2 <= bram_rxpacket_readdata;
						  elsif (bram_rxpacket_readaddress = "000000010") then
						    rxpacket_header_dw3 <= bram_rxpacket_readdata;
						  elsif (bram_rxpacket_readaddress = "000000011") then
						    rxpacket_header_dw4 <= bram_rxpacket_readdata;
						  end if;
						  --decode some parameters: 
						  rxpacket_payloadsize_dwords (7 downto 0) <= rxpacket_header_dw4(25 downto 18); --from bytecount to dword count
						  rxpacket_type <= rxpacket_header_dw4(15 downto 12);
						  rxpacket_status <= rxpacket_header_dw4(11 downto 8);
						  rxpacket_firstdw_be <= rxpacket_header_dw4(7 downto 4);
						  rxpacket_requesterid <= rxpacket_header_dw1(31 downto 16); --same as source id, but only for this type of trn
						  rxpacket_sourceid <= rxpacket_header_dw1(31 downto 16);
						  rxpacket_destinationid <= rxpacket_header_dw1(15 downto 0);
						  rxpacket_targetaddress <= rxpacket_header_dw2(31 downto 0);
						  rxpacket_sourceaddress <= rxpacket_header_dw3(31 downto 0);
						  bram_rxpacket_firstdata_address (8 downto 0)   <= "000000100"; --addr in packet buffer
						  rxpacket_decodedaddress <= rxpacket_targetaddress;
						  bram_txpacket_writeaddress <= "000000100"; --point after the 4dw readcompl header
						  --decide based on header: 
						  -- Packet type: 0000=wr_req, 0001=rd_request, 0010=rd_completion, 0011=forwarding request, 
						  --0100=forw_compl, others: RFU
						  if (packetstm_counter = "00000101") then --when the parameters are already stored
								--read request arrived. (respond with data)
								if (rxpacket_type="0001") then 
									packet_state <= "00000011"; --3
								--read completion arrived. (put data into memory, same as write arrived state)
								elsif (rxpacket_type="0010") then 
									packet_state <= "00000010"; --2
								--write with data arrived
								elsif (rxpacket_type="0000") then 
									packet_state <= "00000010"; --2
								--rd forwarding request came in (put the whole packet into the memory, except at ID match)
								elsif (rxpacket_type="0011") then 
									packet_state <= "00000111"; --7
								--rd forwarding completion came in (someone was reading someone, this is the read data)
								elsif (rxpacket_type="0100") then 
									packet_state <= "00001000"; --8
								--wr forwarding req came in (someone is writing someone with data)
								elsif (rxpacket_type="0101") then 
									packet_state <= "00000111"; --7
								--just wait until this gets a real value
								elsif (rxpacket_type="1111") then 
								 --it also means that the logic will hang here forever if the type=1111.
									packet_state <= "00000001"; --1/decode, stay in this state
								--unsupported request
								else 
								 packet_state <= "00000101";
								end if;
						  end if;

                --********** WRITE arrived STATE **********
					 --initiated by link partner
					 --initiate WB write(s) (1...N DWORD accesses)
                when "00000010" =>   --state 2
						op2p_bar0_wb_addr_o_feed(31 downto 2) <= rxpacket_targetaddress(31 downto 2) + rxpacket_datacount;
						op2p_bar0_wb_addr_o_feed(1 downto 0) <= bit10(1 downto 0);
						op2p_bar0_wb_sel_o_feed  <= rxpacket_firstdw_be;
						op2p_bar0_wb_data_o_feed <= bram_rxpacket_readdata;
						packet_state_copy <= packet_state;
						if (packet_state_copy = packet_state) then 
						  start_write_wb0 <= '0';
						else --generate just one pulse, at the first clk cycle in this state
						  start_write_wb0 <= '1';
						end if;
						if (wb_transaction_complete='1') then  --one DW transfer completed
							bram_rxpacket_readaddress <= bram_rxpacket_readaddress +1;
							rxpacket_datacount <= rxpacket_datacount +1;
							if (rxpacket_payloadsize_dwords = rxpacket_datacount)then  --all data completed
							  packet_state <= "00000000"; --to idle
							else
							  packet_state <= "00010100"; --restart wb transaction with new data
							end if;
						end if;
                --* Write restart state *
                when "00010100" =>   --state 20
 						packet_state <= "00000010";


                --********** READ STATE **********
					 --requested by link partner
					 --initiate WB read, then go to completion state to send data to link partner
                when "00000011" =>   --state 3
						op2p_bar0_wb_addr_o_feed(31 downto 2) <= source_address(31 downto 2) + rxpacket_datacount; --increm by 4
						op2p_bar0_wb_addr_o_feed(1 downto 0) <= bit10(1 downto 0);
						op2p_bar0_wb_sel_o_feed  <= rxpacket_firstdw_be;
						packet_state_copy <= packet_state;
						if (packet_state_copy = packet_state) then 
						  start_read_wb0 <= '0';
						else --generate just one pulse
						  start_read_wb0 <= '1';
						end if;
						if (wb_transaction_complete='1') then 
							bram_txpacket_writedata <= op2p_bar0_wb_data_i_latched; --upd val 
							bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
							rxpacket_datacount <= rxpacket_datacount +1;
							if (rxpacket_payloadsize_dwords = rxpacket_datacount)then 
							  packet_state <= "00000100"; --read completion
							else
							  packet_state <= "00011110"; --one more wb read
							end if;
						end if;
                --* read restart STATE  *
                when "00011110" =>   --state 30
						packet_state <= "00000011";

                --********** READ COMPLETION transmission  STATE **********
					 --requested by link partner (send response with data to link partner)
					 --assemble the tx packet and initiate the transmit
					 --tx packet block ram: bram_txpacket_we, bram_txpacket_writeaddress, bram_txpacket_writedata,
					 --The completion swaps the source/destination ID/address fields!!!!!
                when "00000100" =>   --state 4
                    packet_state_copy <= packet_state;
						  if (bram_txpacket_writeaddress="000000000") then --header 1.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 16) <= rxpacket_header_dw1(15 downto 0); --source ID**
							  bram_txpacket_writedata (15 downto 0) <= rxpacket_header_dw1(31 downto 16); --destination ID**
						  elsif (bram_txpacket_writeaddress="000000001") then --header 2.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 0) <= rxpacket_header_dw3; --target address**
						  elsif (bram_txpacket_writeaddress="000000010") then --header 3.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 0) <= rxpacket_header_dw2;--source address**
						  elsif (bram_txpacket_writeaddress="000000011") then --header 4.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
						     bram_txpacket_writedata (31 downto 16) <= rxpacket_header_dw4 (31 downto 16); --byte count**
							  --packet type: Packet type: 0000=wr_req, 0001=rd_request, 0010=rd_completion, 0011=forwarding request, 0100=forw_compl, others: RFU
								  if (rxpacket_type = "0011") then --read forwarding req
									 bram_txpacket_writedata (15 downto 12) <= "0100";
								  else --normal read req
									 bram_txpacket_writedata (15 downto 12) <= "0010";
								  end if;
							  bram_txpacket_writedata (11 downto 8) <= "0000"; --status** : Status: 0000=succesful_transaction, 0001=no_further_hop, 
							    --0010=unknown_error
							  bram_txpacket_writedata (7 downto 4) <= rxpacket_header_dw4 (7 downto 4); --first byte enable**
							  bram_txpacket_writedata (3 downto 0) <= "0000"; --RFU
						  else --data dwords 
						  --2. read data
							--this was written directly during the read state.
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress;
						    bram_txpacket_we <= "0";
						  end if;
						  --one pulse to start the ep-if statemachine, upon arriving to this state
							if (packet_state_copy = packet_state) then 
							  op2p_there_is_a_new_packet_to_transmit  <= '0';
							else 
							  op2p_there_is_a_new_packet_to_transmit  <= '1';
							end if;
							--back to idle when the ep-if tx is finished: (wait to avoid overwrite)
							if (op2p_packet_tx_complete='1') then 
								packet_state <= "00000000";
							end if;



                --********** forwarding request arrived STATE **********
					 --initiated by link partner.
					 --Check if the destination ID = local ID. if it is, then its a rd/wr request
					 --to this bridge. If its a write then go to the wr state. If it is a read request 
					 --to the local host, then the completion packet MUST BE a forwarding completion packet, 
					 --which results in sending out a read completion packet with type=forwarding_completion.
					 --If the ID does not match, then store packet in RAM, write forw FIFOs and give 
					 --interrupt to local host.
					 --0011=rd_forwarding request, 0100=rd_forw_compl, 0101=wr_forwarding,
                when "00000111" =>   --state 7
						if (rxpacket_header_dw1 (31 downto 16) = aurora_local_id) then --ID-match, so process it here
						  if (rxpacket_type = "0011") then --read
						    packet_state <= "00000011"; --3
						  else --write
						    packet_state <= "00000010"; --2
						  end if;
						else --store it for forwarding
						  if (forw_packetaddr_allocated ='1') then --address should be available
						    packetstm_counter <=  (others => '0');
							 rxpacket_decodedaddress <= fbm_nextraddress;
							 rxpacket_payloadsize_dwords <= rxpacket_payloadsize_dwords + 4; --to copy the header as well
							 bram_rxpacket_readaddress <=  (others => '0'); --to copy the header as well
							 fbm_nextraddress_req <= '0';
							 packet_state <= "00000010"; --2, go to write state with tricked parameters
						  elsif (packetstm_counter = "00000010") then
						    fbm_nextraddress_req <= '0';
							 fbm_actualsize <= rxpacket_payloadsize_dwords(15 downto 0);
							 packetstm_counter <=  packetstm_counter; --stop counting
						  else --when counter is 0 or 1, assert request
						    fbm_nextraddress_req <= '1';
							 fbm_actualsize <= rxpacket_payloadsize_dwords(15 downto 0);
							 packetstm_counter <=  packetstm_counter +1;
						  end if;						  
						end if;


                --********** forwarding completion arrived STATE **********
					 --sent by link partner. 
					 --Check if the destination ID = local ID. if it is, then its a read completion,
					 --so store data in memory, write rdcompletion FIFO. If it is not, store whole 
					 --packet in RAM, write forw FIFO and give interrupt to localhost.
                when "00001000" =>   --state 8
						if (rxpacket_header_dw1 (31 downto 16) = aurora_local_id) then --ID-match, so process it here
						  packet_state <= "00000010"; --2, use the existing write state for this
						else --store it for forwarding 
						  if (forw_packetaddr_allocated='1') then --address should be available
						    packetstm_counter <=  (others => '0');
							 rxpacket_decodedaddress <= fbm_nextraddress;
							 rxpacket_payloadsize_dwords <= rxpacket_payloadsize_dwords +4; --to copy the header as well
							 bram_rxpacket_readaddress <=  (others => '0'); --to copy the header as well
							 fbm_nextraddress_req <= '0';
							 packet_state <= "00000010"; --2, go to write state with tricked parameters
						  elsif (packetstm_counter = "00000010") then
						    fbm_nextraddress_req <= '0';
							 fbm_actualsize <= rxpacket_payloadsize_dwords(15 downto 0);
							 packetstm_counter <=  packetstm_counter; --stop counting
						  else --when counter is 0 or 1, assert request
						    fbm_nextraddress_req <= '1';
							 fbm_actualsize <= rxpacket_payloadsize_dwords(15 downto 0);
							 packetstm_counter <=  packetstm_counter +1;
						  end if;						  
						end if; 



                --********** UNSUPPORTED REQUEST STATE **********
					 --completion response with status=001
                when "00000101" =>   --state 5
                    packet_state_copy <= packet_state;
						  if (bram_txpacket_writeaddress="000000000") then --header 1.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 16) <= rxpacket_header_dw1(15 downto 0); --source ID**
							  bram_txpacket_writedata (15 downto 0) <= rxpacket_header_dw1(31 downto 16); --destination ID**
						  elsif (bram_txpacket_writeaddress="000000001") then --header 2.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 0) <= rxpacket_header_dw3; --target address**
						  elsif (bram_txpacket_writeaddress="000000010") then --header 3.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 0) <= rxpacket_header_dw2;--source address**
						  elsif (bram_txpacket_writeaddress="000000011") then --header 4.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
						     bram_txpacket_writedata (31 downto 16) <= rxpacket_header_dw4 (31 downto 16); --byte count**
							  --packet type: Packet type: 0000=wr_req, 0001=rd_request, 0010=rd_completion, 0011=forwarding request, 0100=forw_compl, others: RFU
								  if (rxpacket_type = "0011") then --read forwarding req
									 bram_txpacket_writedata (15 downto 12) <= "0100";
								  else --normal read req
									 bram_txpacket_writedata (15 downto 12) <= "0010";
								  end if;
							  bram_txpacket_writedata (11 downto 8) <= "0010"; --status** : Status: 0000=succesful_transaction, 0001=no_further_hop, 
							    --0010=unknown_error
							  bram_txpacket_writedata (7 downto 4) <= rxpacket_header_dw4 (7 downto 4); --first byte enable**
							  bram_txpacket_writedata (3 downto 0) <= "0000"; --RFU
						  else --data dwords 
						  --2. read data
							--this was written directly during the read state.
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress;
						    bram_txpacket_we <= "0";
						  end if;
						  --one pulse to start the ep-if statemachine, upon arriving to this state
							if (packet_state_copy = packet_state) then 
							  op2p_there_is_a_new_packet_to_transmit  <= '0';
							else 
							  op2p_there_is_a_new_packet_to_transmit  <= '1';
							end if;
							--back to idle when the ep-if tx is finished: (wait to avoid overwrite)
							if (op2p_packet_tx_complete='1') then 
								packet_state <= "00000000";
							end if;



                --********** outgoing write STATE **********
                --initiated by local host
                --parameter fetch substate
					 when "00010000" =>   --state x
						packetstm_counter <=  packetstm_counter +1;
						if (packetstm_counter = "00000000") then
							packetstm_counter <=  packetstm_counter +1;
							aurora_wr_sourceaddress_regrd_en  <=  '1'; 
							aurora_wr_targetaddress_regrd_en  <=  '1'; 
							aurora_wr_sourceid_regrd_en  <=  '1'; 
							aurora_wr_targetid_regrd_en  <=  '1';
							aurora_wr_bytecount_regrd_en  <=  '1';
						else
							packetstm_counter <=  (OTHERS => '0');
							packet_state <= "00010001"; --to wishbone tr state
							aurora_wr_sourceaddress_regrd_en  <=  '0'; 
							aurora_wr_targetaddress_regrd_en  <=  '0'; 
							aurora_wr_sourceid_regrd_en  <=  '0'; 
							aurora_wr_targetid_regrd_en  <=  '0';
							aurora_wr_bytecount_regrd_en  <=  '0';
							source_address <= aurora_wr_sourceaddress_regdout;
							target_address <= aurora_wr_targetaddress_regdout;
							source_id <= aurora_wr_sourceid_regdout(15 downto 0);
							target_id <= aurora_wr_targetid_regdout(15 downto 0);
							byte_count <= aurora_wr_bytecount_regdout;
							 rxpacket_payloadsize_dwords(29 downto 0) <= aurora_wr_bytecount_regdout(31 downto 2);
							 rxpacket_payloadsize_dwords(31 downto 30) <= "00";
						end if;
					 --wishbone transactions substate
					 when "00010001" =>   --state x 
						op2p_bar0_wb_addr_o_feed(31 downto 2) <= source_address(31 downto 2) + rxpacket_datacount; --increm by 4
						op2p_bar0_wb_addr_o_feed(1 downto 0) <= bit10(1 downto 0);
						op2p_bar0_wb_sel_o_feed  <= "1111";
						packet_state_copy <= packet_state;
						if (packet_state_copy = packet_state) then 
						  start_read_wb0 <= '0';
						else --generate just one pulse
						  start_read_wb0 <= '1';
						end if;
						if (wb_transaction_complete='1') then 
							bram_txpacket_writedata <= op2p_bar0_wb_data_i_latched; --upd val 
							bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
							rxpacket_datacount <= rxpacket_datacount +1;
							if (rxpacket_payloadsize_dwords = rxpacket_datacount)then 
							  packet_state <= "00010011"; --completion
							else
							  packet_state <= "00010010"; --one more wb read
							end if;
						end if;
                --* read restart STATE  *
                when "00010010" =>   --state 30
						packet_state <= "00010001";
					 --packet header assembling substate for outgoing write
					 when "00010011" =>   --state x
                    packet_state_copy <= packet_state;
						  if (bram_txpacket_writeaddress="000000000") then --header 1.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 16) <= source_id; --source ID**
							  bram_txpacket_writedata (15 downto 0) <= target_id; --destination ID**
						  elsif (bram_txpacket_writeaddress="000000001") then --header 2.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 0) <= target_address; --target address**
						  elsif (bram_txpacket_writeaddress="000000010") then --header 3.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 0) <= source_address;--source address**
						  elsif (bram_txpacket_writeaddress="000000011") then --header 4.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
						     bram_txpacket_writedata (31 downto 16) <= byte_count(15 downto 0); --byte count**
							  bram_txpacket_writedata (15 downto 12) <= "0000"; --packet type** : Packet type: 0000=wr_req, 
							    --0001=rd_request, 0010=rd_completion, 0011=forwarding request, 0100=forw_compl, others: RFU
							  bram_txpacket_writedata (11 downto 8) <= "0000"; --status** : Status: 0000=succesful_transaction, 
							    --0001=no_further_hop, 0010=unknown_error
							  bram_txpacket_writedata (7 downto 4) <= "1111"; --first byte enable**
							  bram_txpacket_writedata (3 downto 0) <= "0000"; --RFU
						  else --data dwords 
						  --2. read data
							--this was written directly during the read state.
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress;
						    bram_txpacket_we <= "0";
						  end if;
						  --one pulse to start the ep-if statemachine, upon arriving to this state
							if (packet_state_copy = packet_state) then 
							  op2p_there_is_a_new_packet_to_transmit  <= '0';
							else 
							  op2p_there_is_a_new_packet_to_transmit  <= '1';
							end if;
							--back to idle when the ep-if tx is finished: (wait to avoid overwrite)
							if (op2p_packet_tx_complete='1') then 
								packet_state <= "00000000";
							end if;

                --********** outgoing RD REQ STATE **********
                --initiated by local host
                when "01000000" =>   --state x
						packetstm_counter <=  packetstm_counter +1;
						if (packetstm_counter = "00000000") then
							packetstm_counter <=  packetstm_counter +1;
							aurora_rdreq_localaddress_regrd_en  <=  '1'; 
							aurora_rdreq_targetaddress_regrd_en  <=  '1'; 
							aurora_rdreq_sourceid_regrd_en  <=  '1'; 
							aurora_rdreq_targetid_regrd_en  <=  '1';
							aurora_rdreq_bytecount_regrd_en  <=  '1';
						else
							packetstm_counter <=  (OTHERS => '0');
							packet_state <= "01000001"; --to next state
							aurora_rdreq_localaddress_regrd_en  <=  '0'; 
							aurora_rdreq_targetaddress_regrd_en  <=  '0'; 
							aurora_rdreq_sourceid_regrd_en  <=  '0'; 
							aurora_rdreq_targetid_regrd_en  <=  '0';
							aurora_rdreq_bytecount_regrd_en  <=  '0';
							source_address <= aurora_rdreq_localaddress_regdout;
							target_address <= aurora_rdreq_targetaddress_regdout;
							source_id <= aurora_rdreq_sourceid_regdout(15 downto 0);
							target_id <= aurora_rdreq_targetid_regdout(15 downto 0);
							byte_count <= aurora_rdreq_bytecount_regdout;
							 rxpacket_payloadsize_dwords(29 downto 0) <= aurora_rdreq_bytecount_regdout(31 downto 2);
							 rxpacket_payloadsize_dwords(31 downto 30) <= "00";
						end if;
					 --packet header assembling substate for outgoing RD REQ 
					 when "01000001" =>   --state x
                    packet_state_copy <= packet_state;
						  if (bram_txpacket_writeaddress="000000000") then --header 1.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 16) <= source_id; --source ID**
							  bram_txpacket_writedata (15 downto 0) <= target_id; --destination ID**
						  elsif (bram_txpacket_writeaddress="000000001") then --header 2.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 0) <= target_address; --target address**
						  elsif (bram_txpacket_writeaddress="000000010") then --header 3.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
							  bram_txpacket_writedata (31 downto 0) <= source_address;--source address** --local address for rd req
						  elsif (bram_txpacket_writeaddress="000000011") then --header 4.dw
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress +1;
						    bram_txpacket_we <= "1";
						     bram_txpacket_writedata (31 downto 16) <= byte_count(15 downto 0); --byte count**
							  bram_txpacket_writedata (15 downto 12) <= "0001"; --packet type** : Packet type: 0000=wr_req, 
							    --0001=rd_request, 0010=rd_completion, 0011=forwarding request, 0100=forw_compl, others: RFU
							  bram_txpacket_writedata (11 downto 8) <= "0000"; --status** : Status: 0000=succesful_transaction, 
							    --0001=no_further_hop, 0010=unknown_error
							  bram_txpacket_writedata (7 downto 4) <= "1111"; --first byte enable**
							  bram_txpacket_writedata (3 downto 0) <= "0000"; --RFU
						  else --data dwords 
						  --2. read data
							--this was written directly during the read state.
						    bram_txpacket_writeaddress <= bram_txpacket_writeaddress;
						    bram_txpacket_we <= "0";
						  end if;
						  --one pulse to start the ep-if statemachine, upon arriving to this state
							if (packet_state_copy = packet_state) then 
							  op2p_there_is_a_new_packet_to_transmit  <= '0';
							else 
							  op2p_there_is_a_new_packet_to_transmit  <= '1';
							end if;
							--back to idle when the ep-if tx is finished: (wait to avoid overwrite)
							if (op2p_packet_tx_complete='1') then 
								packet_state <= "00000000";
							end if;



                when others => --error
                      packet_state <= "00000000"; --go to state 0
                end case;     
					 
       end if;        
    end if;
    end process; --end packet statemachine




	--byte enable encoding to bit1:0
	 process ( trn_reset_n, rxpacket_firstdw_be )
    begin
       if (trn_reset_n = '0') then
           bit10(1 downto 0) <="00";
       else
         if (rxpacket_firstdw_be ="0001") then
			  bit10(1 downto 0) <= "00";
         elsif (rxpacket_firstdw_be ="0010") then
			  bit10(1 downto 0) <= "01";
         elsif (rxpacket_firstdw_be ="0100") then
			  bit10(1 downto 0) <= "10";
         elsif (rxpacket_firstdw_be ="1000") then
			  bit10(1 downto 0) <= "11";
         elsif (rxpacket_firstdw_be ="0011") then
			  bit10(1 downto 0) <= "00";
         elsif (rxpacket_firstdw_be ="1100") then
			  bit10(1 downto 0) <= "10";
         elsif (rxpacket_firstdw_be ="1111") then 
			  bit10(1 downto 0) <= "00";
			else --this should never happen
			  bit10(1 downto 0) <= "00";
			end if;
       end if;
    end process;
	 
	 








	-- -----------------------------------------------------------------------------------
	-- INFO TO THE LOCAL HOST ABOUT ARRIVED WRITES AND READ COMPLETIONS ------------------
	-- Basically telling that new data is in the buffer
	
	process (trn_reset_n, user_clk_i) 
    begin
    if (trn_reset_n='0') then 
			aurora_rdcompl_localaddress_regwr_en <=  '0';
			aurora_arrivedwrite_address_regwr_en <=  '0';
			aurora_rdcompl_localaddress_regdin <= (OTHERS => '0');
			aurora_arrivedwrite_address_regdin <= (OTHERS => '0');
    else
      if (user_clk_i'event and user_clk_i = '1') then

			--if the previous state was packet decoding, and now its write-arrived:
			if (packet_state_copy = "00000001" and packet_state = "00000010") then
				if (rxpacket_type="0010") then --read completion arrived
				  aurora_rdcompl_localaddress_regwr_en <=  '1';
				  aurora_arrivedwrite_address_regwr_en <=  '0';
					aurora_rdcompl_localaddress_regdin <= rxpacket_targetaddress;
				elsif (rxpacket_type="0000") then --write data arrived
				  aurora_rdcompl_localaddress_regwr_en <=  '0';
				  aurora_arrivedwrite_address_regwr_en <=  '1';
					aurora_arrivedwrite_address_regdin <= rxpacket_targetaddress;
				else
				  aurora_rdcompl_localaddress_regwr_en <=  '0';
				  aurora_arrivedwrite_address_regwr_en <=  '0';
				end if;
			else
				aurora_rdcompl_localaddress_regwr_en <=  '0';
				aurora_arrivedwrite_address_regwr_en <=  '0';
			end if;   
			
       end if;        
    end if;
    end process;







	 
	-- ----------------------------------------------------------------------------------- 
	--FORWARDING BUFFER MANAGER ----------------------------------------------------------
	--
	--buffer can have max 512 entries, each 1kb buffer, 0.5MBytes total.
	
    process (trn_reset_n, user_clk_i) 
    begin
    if (trn_reset_n='0') then 
       fbm_state <= "00000000"; 
		 aurora_forw_freeupaddr_regrd_en <= '0';
		 aurora_forw_freeupaddr_regrd_en_copy <=  '0';
		 aurora_forw_freeupsize_regrd_en <= '0';
		 forw_packetaddr_allocated <= '0';
		 forwbus_64bit <= (OTHERS => '0');
		 which_64bit_isit <= (OTHERS => '0');
		 forwbus_8bit <= (OTHERS => '0');
		 which_8bit_isit <= (OTHERS => '0');
		 which_bit_isit <= (OTHERS => '0');
			forwbuff_position <= (OTHERS => '0');
			forwbuff_position_byteoffset <= (OTHERS => '0');
			forwbuff_position_bitoffset <= (OTHERS => '0');
			fbm_nextraddress <= (OTHERS => '0');
			forwbuf_descriptor_ddword0 <= (OTHERS => '0');
			forwbuf_descriptor_ddword1 <= (OTHERS => '0');
			forwbuf_descriptor_ddword2 <= (OTHERS => '0');
			forwbuf_descriptor_ddword3 <= (OTHERS => '0');
			forwbuf_descriptor_ddword4 <= (OTHERS => '0');
			forwbuf_descriptor_ddword5 <= (OTHERS => '0');
			forwbuf_descriptor_ddword6 <= (OTHERS => '0');
			forwbuf_descriptor_ddword7 <= (OTHERS => '0');
			aurora_forw_freeupaddr_regrd_en <= '0';
			aurora_forw_freeupaddr_regrd_en_copy <= '0';
			aurora_forw_freeupsize_regrd_en <= '0';
			free_this_location <= (OTHERS => '0');
    else
      if (user_clk_i'event and user_clk_i = '1') then
                case ( fbm_state ) is

                --********** IDLE STATE  **********
					 --Wait for vector requests from the scheduler.
					 --Also wait for new entries in the freeup FIFO. Start reading the value from it here (rd_en=1).
                when "00000000" =>   --state 0        
                  if (fbm_nextraddress_req ='1') then --request from scheduler
						  fbm_state <= "00000001"; 
						elsif (aurora_forw_freeupaddr_regempty='0') then --freeup command from local host
						  fbm_state <= "00000110";
						  aurora_forw_freeupaddr_regrd_en <= '1';
						  aurora_forw_freeupaddr_regrd_en_copy <=  '1';
						  aurora_forw_freeupsize_regrd_en <= '1';
						end if;
						forw_packetaddr_allocated <= '0';

                --********** find step-1 STATE ********** 
                when "00000001" =>   --state 1
						--check 64bit of the descriptor table at once. first clock cycle.
						--if buffer is full, then overwrite ddw7
						if (forwbuf_descriptor_ddword0 /= X"FFFFFFFFFFFFFFFF") then
						  forwbus_64bit <= forwbuf_descriptor_ddword0;
						  which_64bit_isit <= "000";
						elsif (forwbuf_descriptor_ddword1 /= X"FFFFFFFFFFFFFFFF") then
						  forwbus_64bit <= forwbuf_descriptor_ddword1;
						  which_64bit_isit <= "001";
						elsif (forwbuf_descriptor_ddword2 /= X"FFFFFFFFFFFFFFFF") then
						  forwbus_64bit <= forwbuf_descriptor_ddword2;
						  which_64bit_isit <= "010";
						elsif (forwbuf_descriptor_ddword3 /= X"FFFFFFFFFFFFFFFF") then
						  forwbus_64bit <= forwbuf_descriptor_ddword3;
						  which_64bit_isit <= "011";
						elsif (forwbuf_descriptor_ddword4 /= X"FFFFFFFFFFFFFFFF") then
						  forwbus_64bit <= forwbuf_descriptor_ddword4;
						  which_64bit_isit <= "100";
						elsif (forwbuf_descriptor_ddword5 /= X"FFFFFFFFFFFFFFFF") then
						  forwbus_64bit <= forwbuf_descriptor_ddword5;
						  which_64bit_isit <= "101";
						elsif (forwbuf_descriptor_ddword6 /= X"FFFFFFFFFFFFFFFF") then
						  forwbus_64bit <= forwbuf_descriptor_ddword6;
						  which_64bit_isit <= "110";
						else --elsif (forwbuf_descriptor_ddword7 /= X"FFFFFFFFFFFFFFFF") then
						  forwbus_64bit <= forwbuf_descriptor_ddword7;
						  which_64bit_isit <= "111";
						end if;
						fbm_state <= "00000010"; --next
						--fast (ststic timing path) version:
--						if (forwbuf_descriptor_ddword0 = X"FFFFFFFFFFFFFFFF" 
--							 and forwbuf_descriptor_ddword1 = X"FFFFFFFFFFFFFFFF"
--							 and forwbuf_descriptor_ddword2 = X"FFFFFFFFFFFFFFFF"
--							 and forwbuf_descriptor_ddword3 = X"FFFFFFFFFFFFFFFF") then
--							if (forwbuf_descriptor_ddword4 = X"FFFFFFFFFFFFFFFF"
--								 and forwbuf_descriptor_ddword5 = X"FFFFFFFFFFFFFFFF") then
--								if (forwbuf_descriptor_ddword6 = X"FFFFFFFFFFFFFFFF") then
--								  forwbus_64bit <= forwbuf_descriptor_ddword7;
--								  which_64bit_isit <= "111";
--								else
--								  forwbus_64bit <= forwbuf_descriptor_ddword6;
--								  which_64bit_isit <= "110";
--								end if;
--							else
--								if (forwbuf_descriptor_ddword4 = X"FFFFFFFFFFFFFFFF") then
--								  forwbus_64bit <= forwbuf_descriptor_ddword5;
--								  which_64bit_isit <= "101";
--								else
--								  forwbus_64bit <= forwbuf_descriptor_ddword4;
--								  which_64bit_isit <= "100";
--								end if;
--							end if;
--						else
--							if (and forwbuf_descriptor_ddword0 = X"FFFFFFFFFFFFFFFF"
--								 and forwbuf_descriptor_ddword1 = X"FFFFFFFFFFFFFFFF") then
--								if (forwbuf_descriptor_ddword2 = X"FFFFFFFFFFFFFFFF") then
--								  forwbus_64bit <= forwbuf_descriptor_ddword3;
--								  which_64bit_isit <= "011";
--								else
--								  forwbus_64bit <= forwbuf_descriptor_ddword2;
--								  which_64bit_isit <= "010";
--								end if;
--							else
--								if (forwbuf_descriptor_ddword0 = X"FFFFFFFFFFFFFFFF") then
--								  forwbus_64bit <= forwbuf_descriptor_ddword1;
--								  which_64bit_isit <= "001";
--								else
--								  forwbus_64bit <= forwbuf_descriptor_ddword0;
--								  which_64bit_isit <= "000";
--								end if;
--							end if;
--						end if;
						

                --********** find step-2 STATE **********     				 
                when "00000010" =>   --state 2
						--second clock cycle, analyse the 64bit, break it to a final byte:
						--if buffer full, overwrite last byte
						if (forwbuf_descriptor_byte0 /= X"FF") then
						  forwbus_8bit <= forwbuf_descriptor_byte0;
						  which_8bit_isit <= "000";
						elsif (forwbuf_descriptor_byte1 /= X"FF") then
						  forwbus_8bit <= forwbuf_descriptor_byte1;
						  which_8bit_isit <= "001";
						elsif (forwbuf_descriptor_byte2 /= X"FF") then
						  forwbus_8bit <= forwbuf_descriptor_byte2;
						  which_8bit_isit <= "010";
						elsif (forwbuf_descriptor_byte3 /= X"FF") then
						  forwbus_8bit <= forwbuf_descriptor_byte3;
						  which_8bit_isit <= "011";
						elsif (forwbuf_descriptor_byte4 /= X"FF") then
						  forwbus_8bit <= forwbuf_descriptor_byte4;
						  which_8bit_isit <= "100";
						elsif (forwbuf_descriptor_byte5 /= X"FF") then
						  forwbus_8bit <= forwbuf_descriptor_byte5;
						  which_8bit_isit <= "101";
						elsif (forwbuf_descriptor_byte6 /= X"FF") then
						  forwbus_8bit <= forwbuf_descriptor_byte6;
						  which_8bit_isit <= "110";
						else -- elsif (forwbuf_descriptor_byte7 /= X"FF") then
						  forwbus_8bit <= forwbuf_descriptor_byte7;
						  which_8bit_isit <= "111";
						end if;
						fbm_state <= "00000011"; --next
	
                --********** find step-3 STATE **********     				 
                when "00000011" =>   --state 3
						--3rd clock cycle: find a zero bit in the chosen byte: 
						if (forwbuf_descriptor_bit0 = '0') then
						  which_bit_isit <= "000";
						elsif (forwbuf_descriptor_bit1 = '0') then
						  which_bit_isit <= "001";
						elsif (forwbuf_descriptor_bit2 = '0') then
						  which_bit_isit <= "010";
						elsif (forwbuf_descriptor_bit3 = '0') then
						  which_bit_isit <= "011";
						elsif (forwbuf_descriptor_bit4 = '0') then
						  which_bit_isit <= "100";
						elsif (forwbuf_descriptor_bit5 = '0') then
						  which_bit_isit <= "101";
						elsif (forwbuf_descriptor_bit6 = '0') then
						  which_bit_isit <= "110";
						else --elsif(forwbuf_descriptor_bit7 = '0') then
						  which_bit_isit <= "111";
						end if;
						fbm_state <= "00000100"; --next
	
                --********** find step-4 STATE **********     				 
                when "00000100" =>   --state 4
						--4th clock cycle: generate the address with one clock cycle delay, for sufficient timing margin
						forwbuff_position(8 downto 6) <= which_64bit_isit;
						forwbuff_position(5 downto 3) <= which_8bit_isit;
						forwbuff_position(2 downto 0) <= which_bit_isit;
						fbm_state <= "00000101"; --next
						forwbuff_position_byteoffset(18 downto 16) <= which_64bit_isit;
						forwbuff_position_byteoffset(15 downto 13) <= which_8bit_isit;
						forwbuff_position_byteoffset(12 downto 10) <= which_bit_isit;
						forwbuff_position_byteoffset(9 downto 0) <= "0000000000"; --since we use 1kB blocks
						forwbuff_position_bitoffset(5 downto 3) <= which_8bit_isit;
						forwbuff_position_bitoffset(2 downto 0) <= which_bit_isit;
	
                --********** tell the result STATE ********** 
					 --Tell the vector to the scheduler, and set the bit here as used (1)   				 
                when "00000101" =>   --state 5
                    fbm_nextraddress <= forwbuff_position_byteoffset + aurora_forwarding_bufferbase_reg;
						  forw_packetaddr_allocated <= '1';
						  fbm_state <= "00000000"; --back to idle
						  --set the bit as occupied:
						  if    (which_64bit_isit = "000") then
						    forwbuf_descriptor_ddword0 (CONV_INTEGER(forwbuff_position_bitoffset)) <= '1';
						  elsif (which_64bit_isit = "001") then
						    forwbuf_descriptor_ddword1 (CONV_INTEGER(forwbuff_position_bitoffset)) <= '1';
						  elsif (which_64bit_isit = "010") then
						    forwbuf_descriptor_ddword2 (CONV_INTEGER(forwbuff_position_bitoffset)) <= '1';
						  elsif (which_64bit_isit = "011") then
						    forwbuf_descriptor_ddword3 (CONV_INTEGER(forwbuff_position_bitoffset)) <= '1';
						  elsif (which_64bit_isit = "100") then
						    forwbuf_descriptor_ddword4 (CONV_INTEGER(forwbuff_position_bitoffset)) <= '1';
						  elsif (which_64bit_isit = "101") then
						    forwbuf_descriptor_ddword5 (CONV_INTEGER(forwbuff_position_bitoffset)) <= '1';
						  elsif (which_64bit_isit = "110") then
						    forwbuf_descriptor_ddword6 (CONV_INTEGER(forwbuff_position_bitoffset)) <= '1';
						  elsif (which_64bit_isit = "111") then
						    forwbuf_descriptor_ddword7 (CONV_INTEGER(forwbuff_position_bitoffset)) <= '1';
						  end if;


                --********** Freeup STATE ********** 
					 --If the freeup FIFO had a new entry, free up the bit (set to zero)
					 --This state must be 2 clock cycles: 1- calculate bitposition from address value, 2-write bit and go to IDLE				 
                when "00000110" =>   --state 6
                    aurora_forw_freeupaddr_regrd_en <= '0';
						  aurora_forw_freeupaddr_regrd_en_copy <= '0';
						  aurora_forw_freeupsize_regrd_en <= '0';
						  free_this_location <= aurora_forw_freeupaddr_regdout - aurora_forwarding_bufferbase_reg;
						  --aurora_forw_freeupsize_regdout is ignored;
						  if (aurora_forw_freeupaddr_regrd_en_copy='0') then --second clock cycle
							  fbm_state <= "00000000"; --back to idle
							  --set the bit as free:
							  if    (which_64bit_isit_tofreeup = "000") then
								 forwbuf_descriptor_ddword0 (CONV_INTEGER(which_bit_isit_tofreeup)) <= '0';
							  elsif (which_64bit_isit_tofreeup = "001") then
								 forwbuf_descriptor_ddword1 (CONV_INTEGER(which_bit_isit_tofreeup)) <= '0';
							  elsif (which_64bit_isit_tofreeup = "010") then
								 forwbuf_descriptor_ddword2 (CONV_INTEGER(which_bit_isit_tofreeup)) <= '0';
							  elsif (which_64bit_isit_tofreeup = "011") then
								 forwbuf_descriptor_ddword3 (CONV_INTEGER(which_bit_isit_tofreeup)) <= '0';
							  elsif (which_64bit_isit_tofreeup = "100") then
								 forwbuf_descriptor_ddword4 (CONV_INTEGER(which_bit_isit_tofreeup)) <= '0';
							  elsif (which_64bit_isit_tofreeup = "101") then
								 forwbuf_descriptor_ddword5 (CONV_INTEGER(which_bit_isit_tofreeup)) <= '0';
							  elsif (which_64bit_isit_tofreeup = "110") then
								 forwbuf_descriptor_ddword6 (CONV_INTEGER(which_bit_isit_tofreeup)) <= '0';
							  elsif (which_64bit_isit_tofreeup = "111") then
								 forwbuf_descriptor_ddword7 (CONV_INTEGER(which_bit_isit_tofreeup)) <= '0';
							  end if;
						  end if;
						  
                when others => --error
                      fbm_state <= "00000000"; --back to idle
                end case;     
       end if;        
    end if;
    end process;
	 
	--asynchronous: (just renaming some bits, without logic)
	forwbuf_descriptor_byte0 <= forwbus_64bit(7 downto 0);
	forwbuf_descriptor_byte1 <= forwbus_64bit(15 downto 8);
	forwbuf_descriptor_byte2 <= forwbus_64bit(23 downto 16);
	forwbuf_descriptor_byte3 <= forwbus_64bit(31 downto 24);
	forwbuf_descriptor_byte4 <= forwbus_64bit(39 downto 32);
	forwbuf_descriptor_byte5 <= forwbus_64bit(47 downto 40);
	forwbuf_descriptor_byte6 <= forwbus_64bit(55 downto 48);
	forwbuf_descriptor_byte7 <= forwbus_64bit(63 downto 56);
	forwbuf_descriptor_bit0 <= forwbus_8bit(0);
	forwbuf_descriptor_bit1 <= forwbus_8bit(1);
	forwbuf_descriptor_bit2 <= forwbus_8bit(2);
	forwbuf_descriptor_bit3 <= forwbus_8bit(3);
	forwbuf_descriptor_bit4 <= forwbus_8bit(4);
	forwbuf_descriptor_bit5 <= forwbus_8bit(5);
	forwbuf_descriptor_bit6 <= forwbus_8bit(6);
	forwbuf_descriptor_bit7 <= forwbus_8bit(7);
	which_64bit_isit_tofreeup <= free_this_location (18 downto 16);
	which_bit_isit_tofreeup <= free_this_location (15 downto 10);




	-- -----------------------------------------------------------------------------------
	-- INFO TO THE LOCAL HOST ABOUT ARRIVED FORWARDING PACKETS ------------------
	-- Basically telling that new data is in the buffer
	-- do this after passing both forwarding and wb-write states
	
	process (trn_reset_n, user_clk_i) 
    begin
    if (trn_reset_n='0') then 
			aurora_forwreq_pointer_regwr_en  <=  '0';
			aurora_forwreq_pointer_regdin <= (OTHERS => '0');
    else
      if (user_clk_i'event and user_clk_i = '1') then

			--if the previous state was packet decoding, and now its write-arrived:
			if ((packet_state_copy = "00000111" or packet_state_copy = "00001000") and packet_state = "00000010") then
				if (rxpacket_type="0010") then --just started writing date onto wishbone bus
				  aurora_forwreq_pointer_regwr_en <=  '1';
					aurora_forwreq_pointer_regdin <= rxpacket_decodedaddress;
				else
				  aurora_forwreq_pointer_regwr_en <=  '0';
				end if;
			else
				aurora_forwreq_pointer_regwr_en <=  '0';
			end if;   
			
       end if;        
    end if;
    end process;







-- -------- END OF FILE -----------------------------------------------------------
end Behavioral;
